# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import update_wrapper, reduce, partial
import inspect
import operator as op
from typing import (Callable, Generic, Optional, Sequence, Tuple, List, TypeVar,
                    Any)

from jax import core
from jax import linear_util as lu
from jax.tree_util import (tree_flatten, tree_unflatten, tree_map,
                        tree_multimap, treedef_is_leaf, treedef_tuple,
                        register_pytree_node_class)
from jax._src import custom_api_util
from jax._src.util import cache, safe_zip, safe_map, split_list, Unhashable
from jax._src.api_util import flatten_fun_nokwargs, argnums_partial
from jax.core import raise_to_shaped
from jax.errors import UnexpectedTracerError
from jax._src.ad_util import Zero, zeros_like_aval, stop_gradient_p
from jax.interpreters import partial_eval as pe
from jax.interpreters import ad
from jax.interpreters import batching
from jax.interpreters import mlir
from jax.interpreters import xla
from jax.interpreters.batching import not_mapped
from jax.config import config

from jax._src import traceback_util
traceback_util.register_exclusion(__file__)

map = safe_map
zip = safe_zip


### util

def _resolve_kwargs(fun, args, kwargs):
  ba = inspect.signature(fun).bind(*args, **kwargs)
  ba.apply_defaults()
  if ba.kwargs:
    raise TypeError("keyword arguments could not be resolved to positions")
  else:
    return ba.args

def _initial_style_jaxpr(fun, in_avals):
  jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(fun, in_avals)
  return jaxpr, consts

def _close_jaxpr(jaxpr):
  return core.ClosedJaxpr(pe.convert_constvars_jaxpr(jaxpr), ())

def _initial_style_staging() -> bool:
  return core.thread_local_state.trace_state.initial_style

def _sum_tangents(_, x, *xs):
  return reduce(ad.add_tangents, xs, x)

def _zeros_like_pytree(x):
  return tree_map(Zero.from_value, x)

@partial(partial, tree_map)
def _stop_gradient(x):
  if isinstance(x, core.Tracer):
    return stop_gradient_p.bind(x)
  else:
    return x


### JVPs
ReturnValue = TypeVar('ReturnValue')

@custom_api_util.register_custom_decorator_type
class custom_jvp(Generic[ReturnValue]):
  """Set up a JAX-transformable function for a custom JVP rule definition.

  This class is meant to be used as a function decorator. Instances are
  callables that behave similarly to the underlying function to which the
  decorator was applied, except when a differentiation transformation (like
  :py:func:`jax.jvp` or :py:func:`jax.grad`) is applied, in which case a custom
  user-supplied JVP rule function is used instead of tracing into and
  performing automatic differentiation of the underlying function's
  implementation.

  There are two instance methods available for defining the custom JVP rule:
  :py:func:`~jax.custom_jvp.defjvp` for defining a *single* custom JVP rule for
  all the function's inputs, and for convenience
  :py:func:`~jax.custom_jvp.defjvps`, which wraps
  :py:func:`~jax.custom_jvp.defjvp`, and allows you to provide separate
  definitions for the partial derivatives of the function w.r.t. each of its
  arguments.

  For example::

    @jax.custom_jvp
    def f(x, y):
      return jnp.sin(x) * y

    @f.defjvp
    def f_jvp(primals, tangents):
      x, y = primals
      x_dot, y_dot = tangents
      primal_out = f(x, y)
      tangent_out = jnp.cos(x) * x_dot * y + jnp.sin(x) * y_dot
      return primal_out, tangent_out

  For a more detailed introduction, see the tutorial_.

  .. _tutorial: https://jax.readthedocs.io/en/latest/notebooks/Custom_derivative_rules_for_Python_code.html
  """

  def __init__(self,
               fun: Callable[..., ReturnValue],
               nondiff_argnums: Tuple[int, ...] = ()):
    update_wrapper(self, fun)
    self.fun = fun
    self.nondiff_argnums = nondiff_argnums
    self.jvp: Optional[Callable[..., Tuple[ReturnValue, ReturnValue]]] = None

  __getattr__ = custom_api_util.forward_attr

  def defjvp(self, jvp: Callable[..., Tuple[ReturnValue, ReturnValue]]
    ) -> Callable[..., Tuple[ReturnValue, ReturnValue]]:
    """Define a custom JVP rule for the function represented by this instance.

    Args:
      jvp: a Python callable representing the custom JVP rule. When there are no
        ``nondiff_argnums``, the ``jvp`` function should accept two arguments,
        where the first is a tuple of primal inputs and the second is a tuple of
        tangent inputs. The lengths of both tuples are equal to the number of
        parameters of the ``custom_jvp`` function. The ``jvp`` function should
        produce as output a pair where the first element is the primal output
        and the second element is the tangent output. Elements of the input and
        output tuples may be arrays or any nested tuples/lists/dicts thereof.

    Returns:
      None.

    Example::

      @jax.custom_jvp
      def f(x, y):
        return jnp.sin(x) * y

      @f.defjvp
      def f_jvp(primals, tangents):
        x, y = primals
        x_dot, y_dot = tangents
        primal_out = f(x, y)
        tangent_out = jnp.cos(x) * x_dot * y + jnp.sin(x) * y_dot
        return primal_out, tangent_out
    """
    self.jvp = jvp
    return jvp

  def defjvps(self, *jvps: Optional[Callable[..., ReturnValue]]):
    """Convenience wrapper for defining JVPs for each argument separately.

    This convenience wrapper cannot be used together with ``nondiff_argnums``.

    Args:
      *jvps: a sequence of functions, one for each positional argument of the
        ``custom_jvp`` function. Each function takes as arguments the tangent
        value for the corresponding primal input, the primal output, and the
        primal inputs. See the example below.

    Returns:
      None.

    Example::

      @jax.custom_jvp
      def f(x, y):
        return jnp.sin(x) * y

      f.defjvps(lambda x_dot, primal_out, x, y: jnp.cos(x) * x_dot * y,
                lambda y_dot, primal_out, x, y: jnp.sin(x) * y_dot)
    """
    if self.nondiff_argnums:
      raise TypeError("Can't use ``defjvps`` with ``nondiff_argnums``.")

    def jvp(primals, tangents):
      primal_out = self(*primals)
      zeros = _zeros_like_pytree(primal_out)
      all_tangents_out = [jvp(t, primal_out, *primals) if jvp else zeros
                          for t, jvp in zip(tangents, jvps)]
      tangent_out = tree_multimap(_sum_tangents, primal_out, *all_tangents_out)
      return primal_out, tangent_out

    self.defjvp(jvp)

  @traceback_util.api_boundary
  def __call__(self, *args: Any, **kwargs: Any) -> ReturnValue:  # pytype: disable=invalid-annotation
    if not self.jvp:
      msg = "No JVP defined for custom_jvp function {} using defjvp."
      raise AttributeError(msg.format(self.__name__))
    args = _resolve_kwargs(self.fun, args, kwargs)
    if self.nondiff_argnums:
      nondiff_argnums = set(self.nondiff_argnums)
      args = tuple(_stop_gradient(x) if i in nondiff_argnums else x
                   for i, x in enumerate(args))
      diff_argnums = [i for i in range(len(args)) if i not in nondiff_argnums]
      f_, dyn_args = argnums_partial(lu.wrap_init(self.fun), diff_argnums, args,
                                     require_static_args_hashable=False)
      static_args = [args[i] for i in self.nondiff_argnums]
      jvp = _add_args(lu.wrap_init(self.jvp), static_args)
    else:
      f_, dyn_args = lu.wrap_init(self.fun), args
      jvp = lu.wrap_init(self.jvp)
    args_flat, in_tree = tree_flatten(dyn_args)
    flat_fun, out_tree1 = flatten_fun_nokwargs(f_, in_tree)
    flat_jvp, out_tree2 = _flatten_jvp(jvp, in_tree)
    out_flat = custom_jvp_call_p.bind(flat_fun, flat_jvp, *args_flat)
    _, out_tree = lu.merge_linear_aux(out_tree1, out_tree2)
    return tree_unflatten(out_tree, out_flat)

def _add_args(f, extra_args):
  return _add_args_(f, tuple(Unhashable(arg) for arg in extra_args))

@lu.transformation
def _add_args_(extra_args, *args, **kwargs):
  extra_args = tuple([arg.val for arg in extra_args])
  all_args = (extra_args + args)
  yield (yield all_args, kwargs)

@lu.transformation_with_aux
def _flatten_jvp(in_tree, *args):
  primals_in, tangents_in = split_list(args, [len(args) // 2])
  py_primals = tree_unflatten(in_tree, primals_in)
  py_tangents = tree_unflatten(in_tree, tangents_in)
  pair_out = yield (py_primals, py_tangents), {}
  if not isinstance(pair_out, (list, tuple)) or len(pair_out) != 2:
    msg = ("Custom JVP rule must produce a pair (list or tuple of length two) "
           "representing primal and tangent outputs, got {}.")
    raise TypeError(msg.format(pair_out))
  py_primals_out, py_tangents_out = pair_out
  primals_out, out_tree = tree_flatten(py_primals_out)
  tangents_out, out_tree2 = tree_flatten(py_tangents_out)
  if out_tree != out_tree2:
    msg = ("Custom JVP rule must produce primal and tangent outputs with equal "
           "container (pytree) structures, but got {} and {} respectively.")
    raise TypeError(msg.format(out_tree, out_tree2))
  # TODO(mattjj): compare primals' tangent types to tangent objects' types
  primal_avals_out = [
      raise_to_shaped(core.get_aval(x), weak_type=False).strip_named_shape()
      for x in primals_out]
  tangent_avals_out = [
      raise_to_shaped(core.get_aval(t), weak_type=False).strip_named_shape()
      for t in tangents_out]
  if primal_avals_out != tangent_avals_out:
    if len(primal_avals_out) == 1:
      (av1,), (av2,) = primal_avals_out, tangent_avals_out
      msg = ("Custom JVP rule must produce primal and tangent outputs with "
             "equal shapes and dtypes, but got {} and {} respectively.")
      raise TypeError(msg.format(av1.str_short(), av2.str_short()))
    else:
      msg = ("Custom JVP rule must produce primal and tangent outputs with "
             "equal shapes and dtypes, but got:\n{}")
      disagreements = (
          "  primal {} for tangent {}".format(av1.str_short(), av2.str_short())
          for av1, av2 in zip(primal_avals_out, tangent_avals_out) if av1 != av2)
      raise TypeError(msg.format('\n'.join(disagreements)))
  yield primals_out + tangents_out, out_tree

class CustomJVPCallPrimitive(core.CallPrimitive):
  initial_style: core.Primitive

  def bind(self, fun, jvp, *args):
    args = map(core.full_lower, args)
    top_trace = core.find_top_trace(args)
    fun, env_trace_todo1 = process_env_traces(
        fun, self, top_trace and top_trace.level, False)
    jvp, env_trace_todo2 = process_env_traces(
        jvp, self, top_trace and top_trace.level, True)
    tracers = map(top_trace.full_raise, args)  # type: ignore
    outs = top_trace.process_custom_jvp_call(self, fun, jvp, tracers)  # type: ignore
    _, env_trace_todo = lu.merge_linear_aux(env_trace_todo1, env_trace_todo2)
    return _apply_todos(env_trace_todo, map(core.full_lower, outs))

  def impl(self, fun, _, *args):
    with core.new_sublevel():
      return fun.call_wrapped(*args)

  def post_process(self, trace, out_tracers, jvp_was_run: bool):
    return trace.post_process_custom_jvp_call(out_tracers, jvp_was_run)

@lu.transformation_with_aux
def process_env_traces(primitive, level: int, jvp_was_run: bool, *args):
  outs = yield args, {}
  todo = []
  while True:
    tracers = [x for x in outs if isinstance(x, core.Tracer)
               and (level is None or x._trace.level > level)]
    if tracers:
      ans = max(tracers, key=lambda x: x._trace.level)
    else:
      break
    trace = ans._trace.main.with_cur_sublevel()
    outs = map(trace.full_raise, outs)
    outs, cur_todo = primitive.post_process(trace, outs, jvp_was_run)
    todo.append(cur_todo)
  yield outs, tuple(todo)  # Ensure the aux output is immutable

def _apply_todos(todos, outs):
  todos_list = list(todos)
  while todos_list:
    outs = map(core.full_lower, todos_list.pop()(outs))
  return outs

custom_jvp_call_p = CustomJVPCallPrimitive('custom_jvp_call')


def _custom_jvp_call_jaxpr_impl(*args, fun_jaxpr: core.ClosedJaxpr, **params):
  del params  # other params ignored because we're just executing the primal fun
  return core.jaxpr_as_fun(fun_jaxpr)(*args)

def _custom_jvp_call_jaxpr_abstract_eval(*args, fun_jaxpr: core.ClosedJaxpr, **params):
  del args, params
  return fun_jaxpr.out_avals

custom_jvp_call_jaxpr_p = core.AxisPrimitive('custom_jvp_call_jaxpr')
custom_jvp_call_jaxpr_p.multiple_results = True
custom_jvp_call_jaxpr_p.def_impl(_custom_jvp_call_jaxpr_impl)
custom_jvp_call_jaxpr_p.def_abstract_eval(_custom_jvp_call_jaxpr_abstract_eval)
CustomJVPCallPrimitive.initial_style = custom_jvp_call_jaxpr_p

mlir.register_lowering(custom_jvp_call_jaxpr_p, mlir.lower_fun(
    _custom_jvp_call_jaxpr_impl, multiple_results=True))


def _custom_jvp_call_jaxpr_jvp(
    primals, tangents, *, fun_jaxpr: core.ClosedJaxpr,
    jvp_jaxpr_thunk: Callable[[], Tuple[core.Jaxpr, Sequence[Any]]],
    num_consts: int):
  _, args = split_list(primals, [num_consts])
  consts_dot, args_dot = split_list(tangents, [num_consts])
  if any(type(t) is not Zero for t in consts_dot):
    raise ad.CustomJVPException()
  jvp_jaxpr, jvp_consts = jvp_jaxpr_thunk()  # consts can be tracers!
  args_dot = map(ad.instantiate_zeros, args_dot)
  # Cast float0 to zeros with the primal dtype because custom jvp rules don't
  # currently handle float0s
  args_dot = map(ad.replace_float0s, args, args_dot)
  outs = core.eval_jaxpr(jvp_jaxpr, jvp_consts, *args, *args_dot)
  primals_out, tangents_out = split_list(outs, [len(outs) // 2])
  tangents_out = map(ad.recast_to_float0, primals_out, tangents_out)
  return primals_out, tangents_out
ad.primitive_jvps[custom_jvp_call_jaxpr_p] = _custom_jvp_call_jaxpr_jvp

def _custom_jvp_call_jaxpr_vmap(
    axis_size, axis_name, main_type, args, in_dims, *, fun_jaxpr: core.ClosedJaxpr,
    jvp_jaxpr_thunk: Callable[[], Tuple[core.Jaxpr, Sequence[Any]]],
    num_consts: int):
  args = [batching.moveaxis(x, d, 0) if d is not not_mapped and d != 0
          else x for x, d in zip(args, in_dims)]
  num_out = len(fun_jaxpr.out_avals)

  in_batched = [d is not not_mapped for d in in_dims]
  batched_fun_jaxpr, out_batched = batching.batch_jaxpr(
      fun_jaxpr, axis_size, in_batched, False, axis_name, main_type)
  out_dims1 = [0 if b else not_mapped for b in out_batched]
  out_dims2 = []  # mutable cell updated by batched_jvp_jaxpr_thunk

  @pe._memoize
  def batched_jvp_jaxpr_thunk():
    jvp_jaxpr = core.ClosedJaxpr(*jvp_jaxpr_thunk())  # consts can be tracers
    _, args_batched = split_list(in_batched, [num_consts])
    _, all_batched = batching.batch_jaxpr(jvp_jaxpr, axis_size, args_batched * 2, False,
                                          axis_name, main_type)
    primals_batched, tangents_batched = split_list(all_batched, [num_out])
    out_batched = map(op.or_, primals_batched, tangents_batched)
    out_dims2.append([0 if b else not_mapped for b in out_batched])
    batched_jvp_jaxpr, _ = batching.batch_jaxpr(
        jvp_jaxpr, axis_size, args_batched * 2, out_batched * 2,
        axis_name, main_type)
    return batched_jvp_jaxpr.jaxpr, batched_jvp_jaxpr.consts

  batched_outs = custom_jvp_call_jaxpr_p.bind(
      *args, fun_jaxpr=batched_fun_jaxpr,
      jvp_jaxpr_thunk=batched_jvp_jaxpr_thunk, num_consts=num_consts)
  out_dims = out_dims2[0] if out_dims2 else out_dims1
  return batched_outs, out_dims
batching.axis_primitive_batchers[custom_jvp_call_jaxpr_p] = _custom_jvp_call_jaxpr_vmap

xla.register_translation(
    custom_jvp_call_jaxpr_p,
    xla.lower_fun(_custom_jvp_call_jaxpr_impl, new_style=True,
                  multiple_results=True),
    initial_style=True)

# If a (multi)linear function is defined with a custom jvp, then
# custom_jvp_call_jaxpr can appear in jaxprs to be transposed. Since it's
# already been linearized, we can drop the jvp rule.
def _custom_jvp_call_jaxpr_transpose(reduce_axes, cts, *args, fun_jaxpr,
                                     jvp_jaxpr_thunk, num_consts):
  del jvp_jaxpr_thunk, num_consts
  return ad.backward_pass(
      fun_jaxpr.jaxpr, reduce_axes, fun_jaxpr.consts, args, cts)
ad.reducing_transposes[custom_jvp_call_jaxpr_p] = _custom_jvp_call_jaxpr_transpose

def custom_jvp_jaxpr_custom_partial_eval_rule(
    saveable: Callable[..., bool], unks_in: List[bool], inst_in: List[bool],
    eqn: core.JaxprEqn
  ) -> Tuple[Optional[core.JaxprEqn], core.JaxprEqn, List[bool], List[bool], List[core.Var]]:
  # It doesn't make sense to unzip (i.e. break up) a custom_jvp function into
  # constituent parts, so we always perform full remat. An alternative would be
  # to allow the policy function to decide whether the value of a
  # custom_jvp-decorated function's application should be saved or not.
  # TODO(mattjj,jekbradbury): the user writing the custom_jvp-decorated function
  # probably has a better idea for what to do under remat (e.g. if the function
  # contains dots or not), so we should allow for more expressive interaction
  # (e.g. allow the policy to depend on which custom_jvp-decorated function is
  # being applied, or annotating the behavior where custom_vjp is called.)
  inst_out = [True] * len(eqn.outvars)
  new_inst = [x for x, inst in zip(eqn.invars, inst_in)
              if type(x) is core.Var and not inst]
  if any(unks_in):
    unks_out = [True] * len(eqn.outvars)
    return None, eqn, unks_out, inst_out, new_inst
  else:
    unks_out = [False] * len(eqn.outvars)
    return eqn, eqn, unks_out, inst_out, new_inst
pe.partial_eval_jaxpr_custom_rules[custom_jvp_call_jaxpr_p] = \
    custom_jvp_jaxpr_custom_partial_eval_rule  # type: ignore


### VJPs

@custom_api_util.register_custom_decorator_type
class custom_vjp(Generic[ReturnValue]):
  """Set up a JAX-transformable function for a custom VJP rule definition.

  This class is meant to be used as a function decorator. Instances are
  callables that behave similarly to the underlying function to which the
  decorator was applied, except when a reverse-mode differentiation
  transformation (like :py:func:`jax.grad`) is applied, in which case a custom
  user-supplied VJP rule function is used instead of tracing into and performing
  automatic differentiation of the underlying function's implementation. There
  is a single instance method, :py:func:`~jax.custom_vjp.defvjp`, which may be
  used to define the custom VJP rule.

  This decorator precludes the use of forward-mode automatic differentiation.

  For example::

    @jax.custom_vjp
    def f(x, y):
      return jnp.sin(x) * y

    def f_fwd(x, y):
      return f(x, y), (jnp.cos(x), jnp.sin(x), y)

    def f_bwd(res, g):
      cos_x, sin_x, y = res
      return (cos_x * g * y, sin_x * g)

    f.defvjp(f_fwd, f_bwd)

  For a more detailed introduction, see the tutorial_.

  .. _tutorial: https://jax.readthedocs.io/en/latest/notebooks/Custom_derivative_rules_for_Python_code.html
  """

  def __init__(self,
               fun: Callable[..., ReturnValue],
               nondiff_argnums: Tuple[int, ...] = ()):
    update_wrapper(self, fun)
    self.fun = fun
    self.nondiff_argnums = nondiff_argnums
    self.fwd: Optional[Callable[..., Tuple[ReturnValue, Any]]] = None
    self.bwd: Optional[Callable[..., Tuple[Any, ...]]] = None

  __getattr__ = custom_api_util.forward_attr

  def defvjp(self,
             fwd: Callable[..., Tuple[ReturnValue, Any]],
             bwd: Callable[..., Tuple[Any, ...]]) -> None:
    """Define a custom VJP rule for the function represented by this instance.

    Args:
      fwd: a Python callable representing the forward pass of the custom VJP
        rule. When there are no ``nondiff_argnums``, the ``fwd`` function has
        the same input signature as the underlying primal function. It should
        return as output a pair, where the first element represents the primal
        output and the second element represents any "residual" values to store
        from the forward pass for use on the backward pass by the function
        ``bwd``. Input arguments and elements of the output pair may be arrays
        or nested tuples/lists/dicts thereof.
      bwd: a Python callable representing the backward pass of the custom VJP
        rule. When there are no ``nondiff_argnums``, the ``bwd`` function takes
        two arguments, where the first is the "residual" values produced on the
        forward pass by ``fwd``, and the second is the output cotangent with the
        same structure as the primal function output. The output of ``bwd`` must
        be a tuple of length equal to the number of arguments of the primal
        function, and the tuple elements may be arrays or nested
        tuples/lists/dicts thereof so as to match the structure of the primal
        input arguments.

    Returns:
      None.

    Example::

      @jax.custom_vjp
      def f(x, y):
        return jnp.sin(x) * y

      def f_fwd(x, y):
        return f(x, y), (jnp.cos(x), jnp.sin(x), y)

      def f_bwd(res, g):
        cos_x, sin_x, y = res
        return (cos_x * g * y, sin_x * g)

      f.defvjp(f_fwd, f_bwd)
    """
    self.fwd = fwd
    self.bwd = bwd

  @traceback_util.api_boundary
  def __call__(self, *args: Any, **kwargs: Any) -> ReturnValue:  # pytype: disable=invalid-annotation
    if not self.fwd or not self.bwd:
      msg = "No VJP defined for custom_vjp function {} using defvjp."
      raise AttributeError(msg.format(self.__name__))
    args = _resolve_kwargs(self.fun, args, kwargs)
    if self.nondiff_argnums:
      for i in self.nondiff_argnums: _check_for_tracers(args[i])
      nondiff_argnums = set(self.nondiff_argnums)
      dyn_argnums = [i for i in range(len(args)) if i not in nondiff_argnums]
      f_, dyn_args = argnums_partial(lu.wrap_init(self.fun), dyn_argnums, args,
                                     require_static_args_hashable=False)
      static_args = [args[i] for i in self.nondiff_argnums]
      fwd, _ = argnums_partial(lu.wrap_init(self.fwd), dyn_argnums, args,
                               require_static_args_hashable=False)
      bwd = _add_args(lu.wrap_init(self.bwd), static_args)
    else:
      f_, dyn_args = lu.wrap_init(self.fun), args
      fwd, bwd = lu.wrap_init(self.fwd), lu.wrap_init(self.bwd)
    args_flat, in_tree = tree_flatten(dyn_args)
    in_avals = [core.raise_to_shaped(core.get_aval(x)) for x in args_flat]
    flat_fun, out_tree = flatten_fun_nokwargs(f_, in_tree)
    flat_fwd, out_trees = _flatten_fwd(fwd, in_tree)
    flat_bwd = _flatten_bwd(bwd, in_tree, in_avals, out_trees)
    out_flat = custom_vjp_call_p.bind(flat_fun, flat_fwd, flat_bwd, *args_flat,
                                      out_trees=out_trees)
    fst, aux = lu.merge_linear_aux(out_tree, out_trees)
    out_tree = aux if fst else aux[0]
    return tree_unflatten(out_tree, out_flat)

@partial(partial, tree_map)
def _check_for_tracers(x):
  if isinstance(x, core.Tracer):
    msg = ("Found a JAX Tracer object passed as an argument to a custom_vjp "
           "function in a position indicated by nondiff_argnums as "
           "non-differentiable. Tracers cannot be passed as non-differentiable "
           "arguments to custom_vjp functions; instead, nondiff_argnums should "
           "only be used for arguments that can't be or contain JAX tracers, "
           "e.g. function-valued arguments. In particular, array-valued "
           "arguments should typically not be indicated as nondiff_argnums. "
           "\n\n"
           "This behavior recently changed in JAX. "
           "See https://github.com/google/jax/blob/main/docs/custom_vjp_update.md "
           "for more information.")
    raise UnexpectedTracerError(msg)

@lu.transformation_with_aux
def _flatten_fwd(in_tree, *args):
  py_args = tree_unflatten(in_tree, args)
  pair_out = yield py_args, {}
  if not isinstance(pair_out, (list, tuple)) or len(pair_out) != 2:
    msg = ("Custom VJP fwd function must produce a pair (list or tuple of "
           "length two) representing primal outputs and residuals (values "
           "stored from the forward pass for use on the backward pass), "
           "got {}.")
    raise TypeError(msg.format(pair_out))
  py_outs, res = pair_out
  out, out_tree = tree_flatten(py_outs)
  res, res_tree = tree_flatten(res)
  yield res + out, (out_tree, res_tree)

@lu.transformation
def _flatten_bwd(in_tree, in_avals, out_trees, *args):
  out_tree, res_tree = out_trees()
  assert len(args) == res_tree.num_leaves + out_tree.num_leaves
  res, cts_out = split_list(args, [res_tree.num_leaves])
  py_res = tree_unflatten(res_tree, res)
  py_cts_out = tree_unflatten(out_tree, cts_out)
  py_cts_in = yield (py_res, py_cts_out), {}
  # For each None in py_cts_in, indicating an argument for which the rule
  # produces no cotangent, we replace it with a pytree with the structure of the
  # corresponding subtree of in_tree and with leaves of a non-pytree sentinel
  # object, to be replaced with Nones in the final returned result.
  zero = object()  # non-pytree sentinel to replace Nones in py_cts_in
  dummy = tree_unflatten(in_tree, [object()] * in_tree.num_leaves)
  cts_in_flat = []
  append_cts = lambda x, d: cts_in_flat.extend([x] * len(tree_flatten(d)[0]))
  try:
    if not isinstance(py_cts_in, tuple):
      raise ValueError
    tree_multimap(append_cts,
                  tuple(zero if ct is None else ct for ct in py_cts_in), dummy)
  except ValueError:
    _, in_tree2 = tree_flatten(py_cts_in)
    msg = ("Custom VJP rule must produce an output with the same container "
           "(pytree) structure as the args tuple of the primal function, "
           "and in particular must produce a tuple of length equal to the "
           "number of arguments to the primal function, but got VJP output "
           "structure {} for primal input structure {}.")
    raise TypeError(msg.format(in_tree2, in_tree)) from None
  # Ignore any None cotangents, and any corresponding to inputs for which the
  # type doesn't equal the tangent type (i.e. float0s)
  # TODO(mattjj): change this to check if tangent type represents 0dim vspace
  yield [Zero(a.at_least_vspace()) if ct is zero or a != a.at_least_vspace()
         else ct for a, ct in zip(in_avals, cts_in_flat)]


class CustomVJPCallPrimitive(core.CallPrimitive):
  initial_style: core.Primitive

  def bind(self, fun, fwd, bwd, *args, out_trees):
    args = map(core.full_lower, args)
    top_trace = core.find_top_trace(args)
    fun, env_trace_todo1 = process_env_traces(
        fun, self, top_trace and top_trace.level, False)
    fwd, env_trace_todo2 = process_env_traces_fwd(
      fwd, top_trace and top_trace.level, out_trees)
    tracers = map(top_trace.full_raise, args)  # type: ignore
    bwd_ = lu.wrap_init(lambda *args: bwd.call_wrapped(*args))
    outs = top_trace.process_custom_vjp_call(self, fun, fwd, bwd_, tracers,
                                             out_trees=out_trees)
    fst, env_trace_todo = lu.merge_linear_aux(env_trace_todo1, env_trace_todo2)
    if fst:
      return _apply_todos(env_trace_todo, map(core.full_lower, outs))
    else:
      env_trace_todo, bwd_transform = env_trace_todo
      bwd = _apply_bwd_transform(bwd_transform, bwd)
      return _apply_todos(env_trace_todo, map(core.full_lower, outs))

  def impl(self, fun, fwd, bwd, *args, out_trees):
    del fwd, bwd, out_trees
    with core.new_sublevel():
      return fun.call_wrapped(*args)

  def post_process(self, trace, out_tracers, params):
    return trace.post_process_custom_vjp_call(out_tracers, params)
custom_vjp_call_p = CustomVJPCallPrimitive('custom_vjp_call')

@lu.transformation_with_aux
def process_env_traces_fwd(level: int, out_trees, *args):
  outs = yield args, {}
  todo = []
  bwd_transforms = []
  while True:
    tracers = [x for x in outs if isinstance(x, core.Tracer)
               and (level is None or x._trace.level > level)]
    if tracers:
      ans = max(tracers, key=lambda x: x._trace.level)
    else:
      break
    trace = ans._trace.main.with_cur_sublevel()
    outs = map(trace.full_raise, outs)
    outs, cur_todo, bwd_xform = trace.post_process_custom_vjp_call_fwd(outs, out_trees)
    todo.append(cur_todo)
    bwd_transforms.append(bwd_xform)
  yield outs, (tuple(todo), tuple(bwd_transforms))


def _apply_bwd_transform(todos, bwd):
  todos_list = list(todos)
  while todos_list:
    bwd = todos_list.pop()(bwd)
  return bwd

def _custom_vjp_call_jaxpr_impl(*args, fun_jaxpr, **_):
  return core.jaxpr_as_fun(fun_jaxpr)(*args)

def _custom_vjp_call_jaxpr_abstract_eval(*_, fun_jaxpr, **__):
  return fun_jaxpr.out_avals

custom_vjp_call_jaxpr_p = core.AxisPrimitive('custom_vjp_call_jaxpr')
custom_vjp_call_jaxpr_p.multiple_results = True
custom_vjp_call_jaxpr_p.def_impl(_custom_vjp_call_jaxpr_impl)
custom_vjp_call_jaxpr_p.def_abstract_eval(_custom_vjp_call_jaxpr_abstract_eval)
CustomVJPCallPrimitive.initial_style = custom_vjp_call_jaxpr_p

mlir.register_lowering(custom_vjp_call_jaxpr_p, mlir.lower_fun(
    _custom_vjp_call_jaxpr_impl, multiple_results=True))


def _custom_vjp_call_jaxpr_jvp(
    primals, tangents, *, fun_jaxpr: core.ClosedJaxpr,
    fwd_jaxpr_thunk: Callable[[], Tuple[core.Jaxpr, Sequence[Any]]],
    bwd: lu.WrappedFun, out_trees: Callable, num_consts: int):
  _, args = split_list(primals, [num_consts])
  consts_dot, args_dot = split_list(tangents, [num_consts])
  if any(type(t) is not Zero for t in consts_dot):
    raise ad.CustomVJPException()
  fwd_jaxpr, fwd_consts = fwd_jaxpr_thunk()  # consts can be tracers!
  out_tree, res_tree = out_trees()
  args_dot = map(ad.instantiate_zeros, args_dot)
  # Cast float0 to zeros with the primal dtype because custom vjp rules don't
  # currently handle float0s
  args_dot = map(ad.replace_float0s, args, args_dot)
  res_and_primals_out = core.eval_jaxpr(fwd_jaxpr, fwd_consts, *args)
  res, primals_out = split_list(res_and_primals_out, [res_tree.num_leaves])
  avals_out = [raise_to_shaped(core.get_aval(x)) for x in primals_out]
  tangents_out = ad.custom_lin_p.bind(
      *res, *args_dot, num_res=res_tree.num_leaves, bwd=bwd, out_avals=avals_out)
  tangents_out = map(ad.recast_to_float0, primals_out, tangents_out)
  return primals_out, tangents_out
ad.primitive_jvps[custom_vjp_call_jaxpr_p] = _custom_vjp_call_jaxpr_jvp

def _custom_vjp_call_jaxpr_vmap(
    axis_size, axis_name, main_type, args, in_dims, *, fun_jaxpr: core.ClosedJaxpr,
    fwd_jaxpr_thunk: Callable[[], Tuple[core.Jaxpr, Sequence[Any]]],
    bwd: lu.WrappedFun, out_trees: Callable, num_consts: int):
  args = [batching.moveaxis(x, d, 0) if d is not not_mapped and d != 0
          else x for x, d in zip(args, in_dims)]

  in_batched = [d is not not_mapped for d in in_dims]
  _, args_batched = split_list(in_batched, [num_consts])
  batched_fun_jaxpr, out_batched = batching.batch_jaxpr(
      fun_jaxpr, axis_size, in_batched, False, axis_name, main_type)
  out_dims1 = [0 if b else not_mapped for b in out_batched]
  out_dims2 = []

  @pe._memoize
  def batched_fwd_jaxpr_thunk():
    fwd_jaxpr = core.ClosedJaxpr(*fwd_jaxpr_thunk())  # consts can be tracers
    batched_fwd_jaxpr, out_batched = batching.batch_jaxpr(
        fwd_jaxpr, axis_size, args_batched, False, axis_name, main_type)
    out_dims2.append([0 if b else not_mapped for b in out_batched])
    return batched_fwd_jaxpr.jaxpr, batched_fwd_jaxpr.consts

  fwd_args_batched = [0 if b else not_mapped for b in args_batched]
  fwd_out_dims = lambda: out_dims2[0]
  batched_bwd = batching.batch_custom_vjp_bwd(bwd, axis_name, axis_size, fwd_out_dims,
                                              fwd_args_batched, main_type)

  batched_outs = custom_vjp_call_jaxpr_p.bind(
      *args, fun_jaxpr=batched_fun_jaxpr,
      fwd_jaxpr_thunk=batched_fwd_jaxpr_thunk, bwd=batched_bwd,
      out_trees=out_trees, num_consts=num_consts)
  out_dims = out_dims2[0] if out_dims2 else out_dims1
  return batched_outs, out_dims
batching.axis_primitive_batchers[custom_vjp_call_jaxpr_p] = _custom_vjp_call_jaxpr_vmap

xla.register_translation(
    custom_vjp_call_jaxpr_p,
    xla.lower_fun(_custom_vjp_call_jaxpr_impl, new_style=True,
                  multiple_results=True),
    initial_style=True)

batching.primitive_batchers[ad.custom_lin_p] = ad._raise_custom_vjp_error_on_jvp
xla.register_translation(ad.custom_lin_p, ad._raise_custom_vjp_error_on_jvp)

pe.partial_eval_jaxpr_custom_rules[custom_vjp_call_jaxpr_p] = \
    custom_jvp_jaxpr_custom_partial_eval_rule  # type: ignore


def custom_gradient(fun):
  """Convenience function for defining custom VJP rules (aka custom gradients).

  While the canonical way to define custom VJP rules is via ``jax.custom_vjp``,
  the ``custom_gradient`` convenience wrapper follows TensorFlow's
  ``tf.custom_gradient`` API. The difference here is that ``custom_gradient``
  can be used as a decorator on one function that returns both the primal value
  (representing the output of the mathematical function to be differentiated)
  and the VJP (gradient) function. See
  https://www.tensorflow.org/api_docs/python/tf/custom_gradient.

  If the mathematical function to be differentiated has type signature ``a ->
  b``, then the Python callable ``fun`` should have signature
  ``a -> (b, CT b --o CT a)`` where we use ``CT x`` to denote a cotangent type
  for ``x`` and the ``--o`` arrow to denote a linear function. See the example
  below. That is, ``fun`` should return a pair where the first element
  represents the value of the mathematical function to be differentiated and the
  second element is a function to be called on the backward pass of reverse-mode
  automatic differentiation (i.e. the "custom gradient" function).

  The function returned as the second element of the output of ``fun`` can close
  over intermediate values computed when evaluating the function to be
  differentiated. That is, use lexical closure to share work between the forward
  pass and the backward pass of reverse-mode automatic differentiation. However,
  it cannot perform Python control flow which depends on the values of the
  closed-over intermediate values or its cotangent arguments; if the function
  includes such control flow, an error is raised.

  Args:
    fun: a Python callable specifying both the mathematical function to be
      differentiated and its reverse-mode differentiation rule. It should return
      a pair consisting of an output value and a Python callable that represents
      the custom gradient function.

  Returns:
    A Python callable that accepts the same arguments as ``fun`` and returns the
    output value specified by the first element of ``fun``'s output pair.

  For example:

  >>> @jax.custom_gradient
  ... def f(x):
  ...   return x ** 2, lambda g: (g * x,)
  ...
  >>> print(f(3.))
  9.0
  >>> print(jax.grad(f)(3.))
  3.0

  An example with a function on two arguments, so that the VJP function must
  return a tuple of length two:

  >>> @jax.custom_gradient
  ... def f(x, y):
  ...   return x * y, lambda g: (y, x)
  ...
  >>> print(f(3., 4.))
  12.0
  >>> print(jax.grad(f, argnums=(0, 1))(3., 4.))
  (4.0, 3.0)
  """
  @custom_vjp
  def wrapped_fun(*args, **kwargs):
    ans, _ = fun(*args, **kwargs)
    return ans

  def fwd(*args, **kwargs):
    ans, rule = fun(*args, **kwargs)
    ans_flat, out_tree = tree_flatten((ans,))
    rule, in_tree = flatten_fun_nokwargs(lu.wrap_init(rule), out_tree)
    ans_avals = [core.get_aval(x).at_least_vspace() for x in ans_flat]
    jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(rule, ans_avals)
    return ans, Residuals(jaxpr, in_tree(), out_tree, consts)

  def bwd(res, cts):
    jaxpr, in_tree, out_tree, consts = res
    cts_flat, out_tree_ = tree_flatten((cts,))
    if out_tree != out_tree_: raise TypeError(f'{out_tree}\n!=\n{out_tree_}')
    cts_out = core.eval_jaxpr(jaxpr, consts, *cts_flat)
    cts_out = tree_unflatten(in_tree, cts_out)
    if treedef_is_leaf(in_tree):
      cts_out = (cts_out,)
    return cts_out

  wrapped_fun.defvjp(fwd, bwd)
  return wrapped_fun

@register_pytree_node_class
class Residuals:
  def __init__(self, jaxpr, in_tree, out_tree, consts):
    self.jaxpr = jaxpr
    self.in_tree = in_tree
    self.out_tree = out_tree
    self.consts = consts
  def __iter__(self):
    return iter((self.jaxpr, self.in_tree, self.out_tree, self.consts))
  def tree_flatten(self):
    return self.consts, (self.jaxpr, self.in_tree, self.out_tree)
  @classmethod
  def tree_unflatten(cls, aux, consts):
    jaxpr, in_tree, out_tree = aux
    return cls(jaxpr, in_tree, out_tree, consts)


def closure_convert(fun, *example_args):
  """Closure conversion utility, for use with higher-order custom derivatives.

  To define custom derivatives such as with ``jax.custom_vjp(f)``, the target
  function ``f`` must take, as formal arguments, all values involved in
  differentiation. If ``f`` is a higher-order function, in that it accepts as an
  argument a Python function ``g``, then values stored away in ``g``'s closure
  will not be visible to the custom derivative rules, and attempts at AD
  involving these values will fail. One way around this is to convert the
  closure by extracting these values, and to pass them as explicit formal
  arguments across the custom derivative boundary. This utility carries out that
  conversion. More precisely, it closure-converts the function ``fun``
  specialized to the types of the arguments given in ``example_args``.

  When we refer here to "values in the closure" of ``fun``, we do not mean the
  values that are captured by Python directly when ``fun`` is defined (e.g. the
  Python objects in ``fun.__closure__``, if the attribute exists). Rather, we
  mean values encountered during the execution of ``fun`` on ``example_args``
  that determine its output. This may include, for instance, arrays captured
  transitively in Python closures, i.e. in the Python closure of functions
  called by ``fun``, the closures of the functions that they call, and so forth.

  The function ``fun`` must be a pure function.

  Example usage::

    def minimize(objective_fn, x0):
      converted_fn, aux_args = closure_convert(objective_fn, x0)
      return _minimize(converted_fn, x0, *aux_args)

    @partial(custom_vjp, nondiff_argnums=(0,))
    def _minimize(objective_fn, x0, *args):
      z = objective_fn(x0, *args)
      # ... find minimizer x_opt ...
      return x_opt

    def fwd(objective_fn, x0, *args):
      y = _minimize(objective_fn, x0, *args)
      return y, (y, args)

    def rev(objective_fn, res, g):
      y, args = res
      y_bar = g
      # ... custom reverse-mode AD ...
      return x0_bar, *args_bars

    _minimize.defvjp(fwd, rev)

  Args:
    fun: Python callable to be converted. Must be a pure function.
    example_args: Arrays, scalars, or (nested) standard Python
      containers (tuples, lists, dicts, namedtuples, i.e., pytrees)
      thereof, used to determine the types of the formal arguments to
      ``fun``. This type-specialized form of ``fun`` is the function
      that will be closure converted.

  Returns:
    A pair comprising (i) a Python callable, accepting the same
    arguments as ``fun`` followed by arguments corresponding to the
    values hoisted from its closure, and (ii) a list of values hoisted
    from the closure.
  """
  flat_args, in_tree = tree_flatten(example_args)
  in_avals = tuple(map(abstractify, flat_args))
  if config.jax_check_tracer_leaks:
    return _closure_convert_for_avals.__wrapped__(fun, in_tree, in_avals)
  else:
    return _closure_convert_for_avals(fun, in_tree, in_avals)

def _is_perturbed(x: Any) -> bool:
  if isinstance(x, ad.JVPTracer):
    return True
  elif isinstance(x, core.Tracer):
    return any(_is_perturbed(attr) for name, attr in x._contents())
  else:
    return False

@cache()
def _closure_convert_for_avals(fun, in_tree, in_avals):
  wrapped_fun, out_tree = flatten_fun_nokwargs(lu.wrap_init(fun), in_tree)
  jaxpr, out_pvals, consts = pe.trace_to_jaxpr_dynamic(wrapped_fun, in_avals)
  out_tree = out_tree()

  (closure_consts, hoisted_consts), merge = partition_list(_is_perturbed, consts)
  num_consts = len(hoisted_consts)

  def converted_fun(*args_hconsts):
    num_args = len(args_hconsts) - num_consts
    args, hoisted_consts = split_list(args_hconsts, [num_args])
    consts = merge(closure_consts, hoisted_consts)
    all_args, in_tree2 = tree_flatten(tuple(args))
    assert in_tree == in_tree2
    out_flat = core.eval_jaxpr(jaxpr, consts, *all_args)
    return tree_unflatten(out_tree, out_flat)

  return converted_fun, hoisted_consts

def partition_list(choice, lst):
  out = [], []
  which = [out[choice(elt)].append(elt) or choice(elt) for elt in lst]
  def merge(l1, l2):
    i1, i2 = iter(l1), iter(l2)
    return [next(i2 if snd else i1) for snd in which]
  return out, merge

def abstractify(x):
  return core.raise_to_shaped(core.get_aval(x))


### Custom transposition

def linear_call(fun: Callable, fun_transpose: Callable, residual_args,
                linear_args):
  """Call a linear function, with a custom implementation for its transpose.

  The type signatures of ``fun`` and ``fun_transpose`` are:

  .. code-block:: haskell

    fun           :: r -> a -o b
    fun_transpose :: r -> b -o a

  where the ``-o`` arrow indicates a linear function, ``r`` is the
  residual input type and ``a`` is the linear input type.

  The functions ``fun`` and ``fun_transpose`` are coupled as
  transposes of one another. Specifically, the transpose of a
  ``linear_call`` primitive is another ``linear_call`` to
  ``fun_transpose``, with ``fun`` as its custom transposition.

  For example:

  >>> def f(r, x):
  ...   return x / r

  >>> def t(r, t):
  ...   return t / r

  >>> def div_add(x, denom):
  ...   return x + linear_call(f, t, denom, x)

  >>> def transpose(f, x_example):
  ...   def transposed(y):
  ...     x, = jax.linear_transpose(f, x_example)(y)
  ...     return x
  ...   return transposed

  >>> div_add(9., 3.)
  DeviceArray(12., dtype=float32, weak_type=True)

  >>> transpose(partial(div_add, denom=3.), 1.)(18.)  # custom
  DeviceArray(24., dtype=float32, weak_type=True)

  >>> transpose(lambda x: x + x / 3., 1.)(18.)  # reference
  DeviceArray(24., dtype=float32, weak_type=True)

  The above definition of ``f`` illustrates the purpose of a residual
  argument: division is linear in one of its inputs (the dividend
  ``x``) but not the other (the divisor ``r``).

  As another example:

  >>> def custom_id(x):
  ...   def f(_, x): return x
  ...   def t(_, t): return 7.
  ...   return linear_call(f, t, (), x)
  >>> custom_id(1.)
  1.0
  >>> transpose(custom_id, 1.)(1.)
  7.0
  >>> transpose(transpose(custom_id, 1.), 1.)(1.)
  1.0
  >>> transpose(transpose(transpose(custom_id, 1.), 1.), 1.)(1.)
  7.0

  Args:
    fun: a Python callable specifying a linear function. It should
      take two arguments: one of "residual" inputs (type ``r``),
      i.e. inputs in which the function is not necessarly linear, and
      one of "linear" inputs (type ``a``).  It should return output
      whose components are linear in the linear input (type ``b``).
    fun_transpose: a Python callable specifying a structurally linear
      function that is the transpose of ``fun`` with respect to its
      linear inputs. Its first argument is the same residual inputs
      (``r``) as ``fun``. Its second argument is of type
      ``b``. Finally, its output is of type ``a`` and each of its
      component are linear in its second argument (the ``b`` inputs).
    residual_args: Argument in which ``fun`` and ``fun_transpose`` are
      not necessarily linear. Not involved in transposition.
    linear_args: Argument in which ``fun`` and ``fun_transpose`` are
      linear and with respect to which the two are transposes.

  Returns:
    The call result, i.e. ``fun(residual_args, linear_args)``.

  """
  operands_res, res_tree = tree_flatten(residual_args)
  operands_lin, lin_tree = tree_flatten(linear_args)

  f_in_tree = treedef_tuple((res_tree, lin_tree))
  f, out_tree = flatten_fun_nokwargs(lu.wrap_init(fun), f_in_tree)

  res_avals = map(abstractify, operands_res)
  lin_avals = map(abstractify, operands_lin)
  f_jaxpr, f_consts = _initial_style_jaxpr(f, (*res_avals, *lin_avals))
  f_jaxpr = _close_jaxpr(f_jaxpr)
  out_avals = map(core.raise_to_shaped, f_jaxpr.out_avals)

  t_in_tree = treedef_tuple((res_tree, out_tree()))
  t, t_out_tree = flatten_fun_nokwargs(lu.wrap_init(fun_transpose), t_in_tree)

  t_jaxpr, t_consts = _initial_style_jaxpr(t, (*res_avals, *out_avals))
  t_jaxpr = _close_jaxpr(t_jaxpr)

  if t_out_tree() != lin_tree:
    raise TypeError(
        'transpose output pytree structure must match that of linear inputs, '
        f'got output structure {t_out_tree()} '
        f'and input structure {lin_tree}.')

  out = linear_call_p.bind(*f_consts, *t_consts, *operands_res, *operands_lin,
                           callee=f_jaxpr,
                           transpose=t_jaxpr,
                           num_callee_consts=len(f_consts),
                           num_transpose_consts=len(t_consts),
                           num_res=len(operands_res))

  return tree_unflatten(out_tree(), out)

def _linear_call_impl(*args, callee, transpose, num_callee_consts,
                      num_transpose_consts, num_res):
  del transpose
  consts, _, operands_res, operands_lin = split_list(
      args, [num_callee_consts, num_transpose_consts, num_res])
  return core.eval_jaxpr(callee.jaxpr, (), *consts, *operands_res, *operands_lin)

def _linear_call_transpose_rule(cts, *args, callee, transpose,
                                num_callee_consts,
                                num_transpose_consts, num_res):
  f_consts, t_consts, operands_res, operands_lin = split_list(
      args, [num_callee_consts, num_transpose_consts, num_res])
  _, _, cts_avals = split_list(
      transpose.in_avals, [num_transpose_consts, num_res])

  assert all(ad.is_undefined_primal(x)     for x in operands_lin)
  assert all(not ad.is_undefined_primal(x) for x in operands_res)

  cts = [zeros_like_aval(a) if type(ct) is Zero else ct
         for ct, a in zip(cts, cts_avals)]

  cts_out = linear_call_p.bind(*t_consts, *f_consts, *operands_res, *cts,
                               callee=transpose,
                               transpose=callee,
                               num_callee_consts=len(t_consts),
                               num_transpose_consts=len(f_consts),
                               num_res=len(operands_res))

  return [None] * (num_callee_consts + num_transpose_consts + num_res) + cts_out

def _linear_call_abstract_eval(*args, **kwargs):
  return map(core.raise_to_shaped, kwargs['callee'].out_avals)

linear_call_p = core.Primitive('linear_call')
linear_call_p.multiple_results = True
linear_call_p.def_impl(_linear_call_impl)
linear_call_p.def_abstract_eval(_linear_call_abstract_eval)
ad.primitive_transposes[linear_call_p] = _linear_call_transpose_rule
xla.register_translation(linear_call_p,
                         xla.lower_fun(_linear_call_impl, new_style=True,
                                       multiple_results=True),
                         initial_style=True)
mlir.register_lowering(linear_call_p, mlir.lower_fun(
    _linear_call_impl, multiple_results=True))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Demonstrates reuse of a jax2tf model in Keras.

Includes the flags from saved_model_main.py.

See README.md.
"""
import logging
from absl import app
from absl import flags
from jax.experimental.jax2tf.examples import mnist_lib  # type: ignore
from jax.experimental.jax2tf.examples import saved_model_main  # type: ignore
import tensorflow as tf  # type: ignore
import tensorflow_datasets as tfds  # type: ignore
import tensorflow_hub as hub  # type: ignore


FLAGS = flags.FLAGS


def main(_):
  FLAGS.model_classifier_layer = False  # We only need the features
  # Train the model and save the feature extractor
  saved_model_main.train_and_save()

  tf_accelerator, _ = saved_model_main.tf_accelerator_and_tolerances()
  feature_model_dir = saved_model_main.savedmodel_dir()

  # With Keras, we use the tf.distribute.OneDeviceStrategy as the high-level
  # analogue of the tf.device(...) placement seen above.
  # It works on CPU, GPU and TPU.
  # Actual high-performance training would use the appropriately replicated
  # TF Distribution Strategy.
  strategy = tf.distribute.OneDeviceStrategy(tf_accelerator)
  with strategy.scope():
    images = tf.keras.layers.Input(
        mnist_lib.input_shape, batch_size=mnist_lib.train_batch_size)
    # We do not yet train the SavedModel, due to b/123499169.
    keras_feature_extractor = hub.KerasLayer(feature_model_dir, trainable=False)
    features = keras_feature_extractor(images)
    predictor = tf.keras.layers.Dense(10, activation="softmax")
    predictions = predictor(features)
    keras_model = tf.keras.Model(images, predictions)

  keras_model.compile(
      loss=tf.keras.losses.categorical_crossentropy,
      optimizer=tf.keras.optimizers.SGD(learning_rate=0.01),
      metrics=["accuracy"])
  logging.info(keras_model.summary())

  train_ds = mnist_lib.load_mnist(
      tfds.Split.TRAIN, batch_size=mnist_lib.train_batch_size)
  test_ds = mnist_lib.load_mnist(
      tfds.Split.TEST, batch_size=mnist_lib.test_batch_size)
  keras_model.fit(train_ds, epochs=FLAGS.num_epochs, validation_data=test_ds)

  if FLAGS.show_images:
    mnist_lib.plot_images(
        test_ds,
        1,
        5,
        f"Keras inference with reuse of {saved_model_main.model_description()}",
        inference_fn=lambda images: keras_model(tf.convert_to_tensor(images)))


if __name__ == "__main__":
  app.run(main)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa: F401
from jax.scipy import interpolate as interpolate
from jax.scipy import linalg as linalg
from jax.scipy import ndimage as ndimage
from jax.scipy import signal as signal
from jax.scipy import sparse as sparse
from jax.scipy import special as special
from jax.scipy import stats as stats
from jax.scipy import fft as fft



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial, reduce
import itertools as it
import operator as op
from typing import (Tuple, List, Sequence, Set, Dict, Any, Callable, Union,
                    Optional)

from jax import core
from jax._src import dispatch
from jax._src import source_info_util
from jax.core import Var, Literal, Atom, Tracer
from jax._src import util
from jax._src.util import (safe_zip, safe_map, curry, unzip2, split_list,
                           tuple_delete)
import jax._src.pretty_printer as pp

map = safe_map
zip = safe_zip
def identity(x): return x

DType = Any
NDArray = Any


# Dynamic shape jaxprs

## Element types

class EltTy: pass

class BaseType(EltTy):
  def __init__(self, dtype: DType):
    self._dtype = np.dtype(dtype)

  def __repr__(self):
    return f'BaseType({self._dtype.name})'

  def __hash__(self):
    return hash(self._dtype)

  def __eq__(self, other):
    return isinstance(other, BaseType) and self._dtype == other._dtype

class BoundedIntTy(EltTy):
  def __init__(self, bound: int):
    assert isinstance(bound, int)
    self._bound = bound

  def __repr__(self):
    return f'BIntTy{{≤{self._bound}}}'

  def __eq__(self, other):
    return isinstance(other, BoundedIntTy) and self._bound == other._bound


## Array types

class AbsArray(core.AbstractValue):
  def __init__(self, shape, eltTy):
    assert isinstance(shape, tuple)
    assert isinstance(eltTy, EltTy)
    self.shape = shape
    self._eltTy = eltTy

  def str_short(self, short_dtypes=False):
    del short_dtypes  # ignored
    shape = f'[{",".join(str(d) for d in self.shape)}]' if self.shape else ''
    if isinstance(self._eltTy, BoundedIntTy):
      return f'BInt{{≤{self._eltTy._bound}}}{shape}'
    elif isinstance(self._eltTy, BaseType):
      dtype = self._eltTy._dtype.name
      return f'{dtype}{shape}'
    else:
      return repr(self)

  def __eq__(self, other):
    if (isinstance(other, AbsArray) and self._eltTy == other._eltTy and
        len(self.shape) == len(other.shape)):
      for a, b in zip(self.shape, other.shape):
        if type(a) is type(b) is int:
          if a != b: return False
        elif type(a) is type(b) is BoundedInt:
          if a is not b: return False
        elif type(a) is type(b) is Var:
          if a is not b: return False
        elif type(a) is type(b) is AbsArray:
          if a != b: return False
        elif type(a) is type(b) is DimIndexingExpr:
          if a.name is not b.name or a.indices != b.indices: return False
        else:
          return False
      else:
        return True
    return False

  # this duck-typing is needed by eg ad.py using dtypes.py
  @property
  def dtype(self):
    if isinstance(self._eltTy, BaseType):
      return self._eltTy._dtype
    else:
      raise Exception

  def at_least_vspace(self):
    return AbsArray(self.shape, self._eltTy)

  def join(self, other):
    if self == other:
      return self
    raise NotImplementedError  # TODO

class DimIndexingExpr:
  def __init__(self, name, indices):
    assert isinstance(name, (Var, Tracer))
    assert (isinstance(indices, tuple) and
            all(isinstance(i, int) for i in indices))
    self.name = name
    self.indices = indices

  def __repr__(self):
    indices = '.'.join(map(str, self.indices))
    return f'{self.name}.{indices}'


## DJaxprs

class DJaxprTy:
  in_dim_binders: List[Var]
  in_types: List[core.AbstractValue]
  out_dim_binders: List[Var]
  out_types: List[core.AbstractValue]

  def __init__(self, in_dim_binders, in_types, out_dim_binders, out_types):
    self.in_dim_binders = in_dim_binders
    self.in_types = in_types
    self.out_dim_binders = out_dim_binders
    self.out_types = out_types

  def __repr__(self):
    in_dim_binders = pp_vars(self.in_dim_binders)
    in_types = ', '.join(aval.str_short() for aval in self.in_types)
    out_dim_binders = pp_vars(self.out_dim_binders)
    out_types = ', '.join(aval.str_short() for aval in self.out_types)
    return f'[{in_dim_binders}] [{in_types}] -> [{out_dim_binders}] [{out_types}]'

class DJaxpr:
  in_dim_binders: List[Var]
  in_binders: List[Var]
  out_dims: List[Atom]
  outs: List[Atom]
  eqns: List[core.JaxprEqn]  # reusing existing eqns, helps reuse some tracing

  def __init__(self, in_dim_binders, in_binders, out_dims, outs, eqns):
    assert all(isinstance(v, Var) and isinstance(v.aval, AbsArray) and
               isinstance(v.aval._eltTy, BoundedIntTy) for v in in_dim_binders)
    assert all(isinstance(v, Var) for v in in_binders)
    assert all(isinstance(x, (Var, Literal)) and isinstance(x.aval, AbsArray) and
               isinstance(x.aval._eltTy, BoundedIntTy) for x in out_dims)
    assert all(isinstance(x, (Var, Literal)) for x in outs)
    assert all(isinstance(e, core.JaxprEqn) for e in eqns)
    self.in_dim_binders = in_dim_binders
    self.in_binders = in_binders
    self.out_dims = out_dims
    self.outs = outs
    self.eqns = eqns

  def __repr__(self):
    return str(pp_djaxpr(self))

def pp_djaxpr(jaxpr: DJaxpr) -> pp.Doc:
  eqns = map(pp_eqn, jaxpr.eqns)
  in_dim_binders = pp_vars(jaxpr.in_dim_binders)
  in_binders = pp_vars(jaxpr.in_binders)
  out_dims = ', '.join(map(str, jaxpr.out_dims))
  outs = ', '.join(map(str, jaxpr.outs))
  out_dim_types = pp_vars(jaxpr.out_dims)
  outs_type = ', '.join(v.aval.str_short() for v in jaxpr.outs)
  return (pp.text(f'{{ lambda {in_dim_binders} ; {in_binders} .')
          + (pp.text('let ') + pp.nest(2, pp.brk() + pp.join(pp.brk(), eqns)) +
             pp.text(f'in ( {out_dims} ; {outs} ) '
                f': ( {out_dim_types} ; {outs_type} ) }}')))

def pp_vars(vs: Sequence[Atom]) -> str:
  return ', '.join(f'{v}:{v.aval.str_short()}' for v in vs)

def pp_eqn(eqn: core.JaxprEqn) -> pp.Doc:
  lhs = pp_vars(eqn.outvars)
  pp_lhs = pp.text(f'{lhs} =')
  pp_rhs = (pp.text(eqn.primitive.name) +
            core.pp_kv_pairs(sorted(eqn.params.items()), core.JaxprPpContext())
            + pp.text(' ') + pp.text(' '.join(map(str, eqn.invars))))
  return pp_lhs + pp.text(' ') + pp_rhs

# Typechecking DJaxprs

def typecheck_jaxpr(jaxpr: DJaxpr):
  env: Set[Var] = set()  # bound variables

  for v in jaxpr.in_dim_binders:
    if not (isinstance(v.aval, AbsArray) and
            isinstance(v.aval._eltTy, BoundedIntTy)): raise TypeError
    typecheck_type(env, v.aval)
    env.add(v)

  for v in jaxpr.in_binders:
    typecheck_type(env, v.aval)
  for v in jaxpr.in_binders:
    env.add(v)

  for eqn in jaxpr.eqns:
    for x in eqn.invars:
      typecheck_atom(env, x)
    rule = typecheck_rules[eqn.primitive]
    out_types = rule(*eqn.invars, **eqn.params)
    subst: Dict[Var, Var] = {}
    for v, t in zip(eqn.outvars, out_types):
      if isinstance(t, Var):
        aval = substitute(subst, t.aval)
        if v.aval != aval: raise TypeError(f'{v.aval} != {aval}')
        subst[t] = v
      elif isinstance(t, core.AbstractValue):
        aval = substitute(subst, t)
        if v.aval.strip_weak_type() != aval:
          raise TypeError(f'{v.aval} != {aval}')
      else:
        assert False  # typecheck rule produced unexpected type
      typecheck_type(env, v.aval)
      env.add(v)

  in_types = [v.aval for v in jaxpr.in_binders]
  out_types = []
  for x in jaxpr.outs:
    aval = typecheck_atom(env, x)
    out_types.append(aval)

  return DJaxprTy(jaxpr.in_dim_binders, in_types, jaxpr.out_dims, out_types)

def typecheck_type(env, aval):
  if isinstance(aval, (core.AbstractUnit, core.ShapedArray)):
    return aval  # all syntactic forms are valid
  elif isinstance(aval, AbsArray):
    for i, d in enumerate(aval.shape):
      if isinstance(d, int):
        continue
      elif isinstance(d, Var):
        if d not in env: raise TypeError('unbound dim size')
        if not (isinstance(d.aval, AbsArray) and not d.aval.shape and
                isinstance(d.aval._eltTy, BoundedIntTy)):
          raise TypeError(f'dim var of unexpected type: {d.aval}')
      elif isinstance(d, DimIndexingExpr):
        if d.name not in env: raise TypeError('unbound dim size')
        if not (isinstance(d.name.aval, AbsArray) and
                isinstance(d.name.aval._eltTy, BoundedIntTy)):
          raise TypeError(f'dim var of unexpected type: {d.name.aval}')
        d_indices_set = set(d.indices)
        if i in d_indices_set:
          raise TypeError(f"circular dim indexing expression: {d}")
        for j in d.indices:
          d_j = aval.shape[j]
          if (isinstance(d_j, DimIndexingExpr) and
              not d_indices_set.issuperset(d_j.indices)):
            raise TypeError(f"dim indexing not transitively closed: {d}")
        expected_idx_array_shape = tuple(aval.shape[j] for j in d.indices)
        if d.name.aval.shape != expected_idx_array_shape:
          raise TypeError(f'incompatible shapes in dim indexing: {aval}')
      else:
        raise TypeError(f'unexpected type in shape: {type(d)}')
    return aval
  else:
    raise TypeError(f'unknown type: {aval}')

def typecheck_atom(env, x):
  if isinstance(x, Literal):
    return core.raise_to_shaped(core.get_aval(x.val))
  elif isinstance(x, Var):
    return typecheck_type(env, x.aval)
  else:
    raise TypeError(f'atom of unexpected type {x}')

def substitute(subst, aval):
  if isinstance(aval, AbsArray):
    new_shape = []
    for d in aval.shape:
      if isinstance(d, Var):
        new_d = subst.get(d, d)
      elif isinstance(d, DimIndexingExpr):
        new_d = DimIndexingExpr(subst.get(d.name, d.name), d.indices)
      else:
        new_d = d
      new_shape.append(new_d)
    return AbsArray(tuple(new_shape), aval._eltTy)
  else:
    return aval

typecheck_rules: Dict[core.Primitive, Callable] = {}


# Interpreting DJaxprs

def eval_jaxpr(jaxpr, dim_args, args):
  env: Dict[Var, Any] = {}

  def read(v):
    if type(v) is core.Literal:
      return v.val
    else:
      return env[v]

  def write(v, val):
    env[v] = val

  write(core.unitvar, core.unit)
  map(write, jaxpr.in_dim_binders, dim_args)
  map(write, jaxpr.in_binders, args)
  for eqn in jaxpr.eqns:
    in_vals = map(read, eqn.invars)
    ans = eqn.primitive.bind(*in_vals, **eqn.params)
    if eqn.primitive.multiple_results:
      map(write, eqn.outvars, ans)
    elif len(eqn.outvars) > 1:
      # TODO a jaxpr unpacks dependent tuples, while Python packages them up
      map(write, eqn.outvars, eqn.primitive.unpack_result(ans))
    else:
      write(eqn.outvars[0], ans)
  return map(read, jaxpr.out_dims), map(read, jaxpr.outs)

@curry
def jaxpr_as_fun(jaxpr, *args):
  shapevars_to_vals: Dict[Var, Any] = dict(
      (d, t) for v, x in zip(jaxpr.in_binders, args) if isinstance(v.aval, AbsArray)
      for d, t in zip(v.aval.shape, x.shape) if isinstance(d, Var)
      and x is not core.unit)  # TODO partial eval assumes we can plug in units?
  dim_args = [shapevars_to_vals[v] for v in jaxpr.in_dim_binders]
  _, out = eval_jaxpr(jaxpr, dim_args, args)
  return out


# Data representations

class BoundedInt:
  val: Union[int, Tracer]
  bound: int

  def __init__(self, val: Union[int, Tracer], bound: int):
    self._val = val
    self._bound = bound

  def __repr__(self):
    return f'{self._val}{{≤{self._bound}}}'

  def __eq__(self, other):
    if isinstance(other, BoundedInt) and self._bound == other._bound:
      return self._val is other._val or self._val == other._val
    elif isinstance(other, int):
      return self._val == other
    else:
      raise Exception

class DimIndexer:
  data: NDArray
  indices: Tuple[int, ...]

  def __init__(self, data, indices):
    self._data = data
    self._indices = indices

  def __repr__(self):
    indices = '.'.join(map(str, self._indices))
    data = f'{self._data._data}'
    return f'{data}.{indices}'

# We want these to duck-type ndarrays when the element type is BaseType.
class Array:
  def __init__(self,
               shape: Tuple[Union[int, BoundedInt, DimIndexer], ...],
               eltTy: EltTy,
               data: NDArray):
    self.shape = shape
    self._eltTy = eltTy
    self._data = data

  @property
  def dtype(self):
    if isinstance(self._eltTy, BaseType):
      return self._eltTy._dtype
    else:
      raise Exception

  def __repr__(self):
    dtypestr = (self._eltTy._dtype.name if isinstance(self._eltTy, BaseType)
                else f'BInt{{≤{self._eltTy._bound}}}')  # type: ignore
    shapestr = ','.join(map(str, self.shape))
    if any(isinstance(d, DimIndexer) for d in self.shape):
      # find the last DimIndexer, as we'll treat chunks below that as
      # rectangular
      last = next(i for i, d in reversed(list(enumerate(self.shape)))
                  if isinstance(d, DimIndexer))
      shape_prefix = tuple(d._val if type(d) is BoundedInt else d
                            for d in self.shape[:last])
      outs = []
      for idx in it.product(*map(range, shape_prefix)):
        slices = [slice(d._data._data[tuple(idx[i] for i in d._indices)])
                  if isinstance(d, DimIndexer) else
                  slice(d._val) if isinstance(d, BoundedInt) else
                  slice(None) for d in self.shape[last:]]
        full_index = (*idx, *slices)
        data = self._data[full_index]
        outs.append(f'{idx}:\n{data}')
      return f'{dtypestr}[{shapestr}] with values:\n' + '\n\n'.join(outs)
    else:
      slices = tuple(slice(d._val) if type(d) is BoundedInt else slice(None)
                     for d in self.shape)
      data = self._data[slices]
      return f'{dtypestr}[{shapestr}] with value:\n{data}'

  def __array__(self):
    if any(isinstance(d, DimIndexer) for d in self.shape):
      raise NotImplementedError  # ragged ndarray
    else:
      slices = tuple(slice(d._val) if type(d) is BoundedInt else slice(None)
                     for d in self.shape)
      return np.array(self._data[slices])


# Tracing to embed DJaxprs in Python

from jax import linear_util as lu
from jax.interpreters import partial_eval as pe

from jax._src.api_util import flatten_fun
from jax.tree_util import tree_flatten, tree_unflatten

def make_djaxpr(fun, *args, **kwargs):
  args, in_tree = tree_flatten((args, kwargs))
  f, out_tree = flatten_fun(lu.wrap_init(fun), in_tree)
  in_avals = [core.raise_to_shaped(core.get_aval(x)) for x in args]
  return trace_to_jaxpr_dynamic(f, in_avals)

def trace_to_jaxpr_dynamic(fun: lu.WrappedFun, in_avals: Sequence[core.AbstractValue]):
  with core.new_main(DJaxprTrace, dynamic=True) as main:
    main.jaxpr_stack = ()  # type: ignore
    outs = trace_to_subjaxpr_dynamic(fun, main, in_avals)
    del main
  return outs

def trace_to_subjaxpr_dynamic(fun: lu.WrappedFun, main: core.MainTrace,
                              in_avals: Sequence[core.AbstractValue]):
  frame = DJaxprStackFrame()
  with pe.extend_jaxpr_stack(main, frame):
    trace = DJaxprTrace(main, core.cur_sublevel())
    in_dim_tracers, in_avals = _place_in_dim_tracers_in_shapes(trace, in_avals)
    in_tracers = map(trace.new_arg, in_avals)
    ans = fun.call_wrapped(*in_tracers)
    out_tracers = map(trace.full_raise, ans)
  out_dim_tracers = _extract_out_dim_tracers_from_shapes(main, in_dim_tracers, out_tracers)
  return frame.to_jaxpr(in_dim_tracers, in_tracers, out_dim_tracers, out_tracers)

def _place_in_dim_tracers_in_shapes(trace, in_avals):
  dim_tracers = {}
  new_in_avals = []
  for aval in in_avals:
    if not isinstance(aval, AbsArray):
      new_in_avals.append(aval)
    else:
      new_shape = []
      for d in aval.shape:
        if isinstance(d, AbsArray):
          assert d.shape == () and isinstance(d._eltTy, BoundedIntTy)
          dim_tracer = dim_tracers.get(id(d))
          if dim_tracer is None:
            dim_tracer = dim_tracers[id(d)] = trace.new_arg(d)
          new_shape.append(dim_tracer)
        elif isinstance(d, (int, BoundedInt)):
          new_shape.append(d)
        else:
          raise NotImplementedError(d)  # TODO
      new_aval = AbsArray(tuple(new_shape), aval._eltTy)
      new_in_avals.append(new_aval)
  return list(dim_tracers.values()), new_in_avals

def _extract_out_dim_tracers_from_shapes(main, in_dim_tracers, out_tracers):
  seen = {id(d) for d in in_dim_tracers}
  def take(d):
    if isinstance(d, Tracer):
      return d._trace.main is main and id(d) not in seen and not seen.add(id(d))
    elif isinstance(d, DimIndexingExpr):
      return take(d.name)
    else:
      return False
  return [d.name if isinstance(d, DimIndexingExpr) else d
          for t in out_tracers if isinstance(t.aval, AbsArray)
          for d in t.aval.shape if take(d)]

class DJaxprTrace(pe.DynamicJaxprTrace):
  def process_primitive(self, primitive, tracers, params):
    rule = custom_staging_rules.get(primitive)
    if rule:
      return rule(self, tracers, params)
    else:
      # If there's no special staging rule, by default do regular Jaxpr staging
      return super().process_primitive(primitive, tracers, params)

  def get_const(self, tracer):
    assert isinstance(tracer, Tracer)
    return self.frame.constvar_to_val.get(self.frame.tracer_to_var.get(id(tracer)))

  def new_const(self, val):
    if isinstance(val, BoundedInt):
      raise NotImplementedError  # TODO
    elif isinstance(val, Array) and val.shape:
      raise NotImplementedError  # TODO
    else:
      return super().new_const(val)

custom_staging_rules: Dict[core.Primitive, Callable] = {}

class DJaxprStackFrame(pe.JaxprStackFrame):
  def to_jaxpr(self, in_dim_tracers, in_tracers, out_dim_tracers, out_tracers):
    t2v = lambda t: self.tracer_to_var[id(t)]
    in_dim_binders, in_binders = map(t2v, in_dim_tracers), map(t2v, in_tracers)
    out_dims, outs = map(t2v, out_dim_tracers), map(t2v, out_tracers)

    # only include constants that are used
    used_vars = ({a for eqn in self.eqns for a in eqn.invars if isinstance(a, Var)} |
                 {a for grp in [out_dims, outs] for a in grp if isinstance(a, Var)})
    constvars, constvals = unzip2(
        (v, c) for v, c in self.constvar_to_val.items() if v in used_vars)
    in_binders = [*constvars, *in_binders]

    # promote some lambda binders to pi binders
    used_shape_vars = ({d for eqn in self.eqns for v in eqn.outvars
                        if isinstance(v.aval, AbsArray)
                        for d in v.aval.shape if isinstance(d, Var)} |
                       {d.name for eqn in self.eqns for v in eqn.outvars
                        if isinstance(v.aval, AbsArray)
                        for d in v.aval.shape if isinstance(d, DimIndexingExpr)})
    lambda_binders = [v not in used_shape_vars for v in in_binders]
    converted_binders, in_binders = partition_list(lambda_binders, in_binders)
    in_dim_binders = in_dim_binders + converted_binders
    out_dims = [v for v in out_dims if v not in in_dim_binders]  # TODO

    jaxpr = DJaxpr(in_dim_binders, in_binders, out_dims, outs, self.eqns)
    typecheck_jaxpr(jaxpr)
    return jaxpr, constvals, lambda_binders

  def newvar(self, aval):
    if isinstance(aval, AbsArray) and aval.shape:
      # replace any tracers in the shape with their corresponding variables
      shape = []
      for d in aval.shape:
        if isinstance(d, Tracer):
          shape.append(self.tracer_to_var[id(d)])
        elif isinstance(d, DimIndexingExpr):
          assert isinstance(d.name, Tracer)
          shape.append(DimIndexingExpr(self.tracer_to_var[id(d.name)], d.indices))
        else:
          shape.append(d)
      aval = AbsArray(tuple(shape), aval._eltTy)
    return self.gensym(aval)

def partition_list(bs, lst):
  lists = lst1, lst2 = [], []
  for b, x in zip(bs, lst):
    lists[b].append(x)
  return lst1, lst2

def _raise_absarray_to_type_level(aval: AbsArray, weak_type: bool):
  assert isinstance(aval, AbsArray)
  unique_avals: Dict[int, AbsArray] = {}
  shape = []
  for d in aval.shape:
    if isinstance(d, BoundedInt):
      shape.append(unique_avals.setdefault(id(d), AbsArray((), BoundedIntTy(d._bound))))
    elif isinstance(d, DimIndexer):
      raise NotImplementedError  # TODO
    else:
      shape.append(d)
  return AbsArray(tuple(shape), aval._eltTy)
core.raise_to_shaped_mappings[AbsArray] = _raise_absarray_to_type_level

def _abstractify_array_for_ad(x: Array):  # TODO misleading name, used in djit
  return AbsArray(x.shape, x._eltTy)
core.pytype_aval_mappings[Array] = _abstractify_array_for_ad

def _abstractify_bdint(x: BoundedInt):
  return AbsArray((), BoundedIntTy(x._bound))
core.pytype_aval_mappings[BoundedInt] = _abstractify_bdint


# XLA lowering

from jax.interpreters import xla
from jax._src.lib import xla_bridge as xb
from jax._src.lib import xla_client as xc
xe = xc._xla
xops = xc._xla.ops

def _abstractify_array_to_type_level(x: Array):
  return core.raise_to_shaped(core.get_aval(x))
xla.pytype_aval_mappings[Array] = _abstractify_array_to_type_level

def _array_xla_shape(aval: AbsArray):
  if isinstance(aval._eltTy, BaseType):
    dtype = aval._eltTy._dtype
    shape = [d._eltTy._bound if isinstance(d, AbsArray) and not d.shape
             else d for d in aval.shape]
    return (xla.xc.Shape.array_shape(dtype, shape),)
  elif isinstance(aval._eltTy, BoundedIntTy):
    shape = [d._bound if isinstance(d, BoundedInt) else d for d in aval.shape]
    return (xla.xc.Shape.array_shape(np.dtype('int32'), shape),)
  else:
    raise NotImplementedError
xla.xla_shape_handlers[AbsArray] = _array_xla_shape
xla.canonicalize_dtype_handlers[Array] = identity

def _array_device_put(x, device):
  return dispatch._device_put_array(x._data, device)
dispatch.device_put_handlers[Array] = _array_device_put

def _bdint_device_put(x, device):
  return dispatch._device_put_scalar(x._val, device)
dispatch.device_put_handlers[BoundedInt] = _bdint_device_put

def _bdint_canoncalize_dtype(x):
  return BoundedInt(xla.canonicalize_dtype(x._val), x._bound)
xla.canonicalize_dtype_handlers[BoundedInt] = _bdint_canoncalize_dtype

def _make_params(c, dim_in_avals, in_avals):
  n = it.count()
  make = lambda a: [xla.parameter(c, next(n), s) for s in xla.aval_to_xla_shapes(a)]
  return map(make, dim_in_avals), map(make, in_avals)

def _xla_consts(c, consts):
  unique_consts = {id(const): const for const in consts}
  xla_consts = {
      id_: [xla.pyval_to_ir_constant(c, const)]
      for id_, const in unique_consts.items()}
  return [xla_consts[id(const)] for const in consts]

def djaxpr_subcomp(c, jaxpr, dim_args, args):
  env: Dict[Var, Sequence[xe.XlaOp]] = {}

  def aval(v):
    return xla.abstractify(v.val) if type(v) is core.Literal else v.aval

  def read(v):
    if type(v) is core.Literal:
      return [xla.pyval_to_ir_constant(c, xla.canonicalize_dtype(v.val))]
    else:
      return env[v]

  def write(v, nodes):
    env[v] = nodes

  write(core.unitvar, xla._make_unit_constant(c))
  map(write, jaxpr.in_dim_binders, dim_args)
  map(write, jaxpr.in_binders, args)
  for eqn in jaxpr.eqns:
    in_vals, in_avals = map(read, eqn.invars), map(aval, eqn.invars)
    in_dims = {v:read(v) for a in in_avals if isinstance(a, AbsArray)
               for v in a.shape if isinstance(v, Var)}
    rule = translations[eqn.primitive]
    out_vals = rule(c, in_dims, in_avals, in_vals, **eqn.params)
    map(write, eqn.outvars, out_vals)
  return map(read, jaxpr.out_dims), map(read, jaxpr.outs)

def execute_compiled(compiled, partitioner, handlers, dim_vals, args):
  input_bufs = list(it.chain(
      (buf for x in dim_vals for buf in dispatch.device_put(x, None)),
      (buf for x in args     for buf in dispatch.device_put(x, None))))
  out_bufs = compiled.execute(input_bufs)
  dims_dict, grouped_bufs = partitioner(out_bufs)
  return [handler(dims_dict, bs) for handler, bs in zip(handlers, grouped_bufs)]

def result_partitioner(in_dim_binders, in_dim_vals, out_dims, out_bufcounts):
  out_dimvars = [v for v in out_dims if isinstance(v, Var)]
  split_sizes = [len(out_dimvars)] + out_bufcounts[:-1]

  def dim_handler(v, buf):
    if not v.aval.shape:
      return BoundedInt(int(buf.to_py()), v.aval._eltTy._bound)
    else:
      return Array(v.aval.shape, v.aval._eltTy, buf.to_py())

  def partitioner(bufs):
    dim_bufs, *grouped_bufs = split_list(bufs, split_sizes)
    dims_dict = dict(it.chain(
        zip(in_dim_binders, in_dim_vals),
        zip(out_dimvars, map(dim_handler, out_dimvars, dim_bufs))))
    return dims_dict, grouped_bufs
  return partitioner

def result_handler(aval):
  if isinstance(aval, AbsArray):
    return array_result_handler(aval)
  else:
    handler = dispatch.aval_to_result_handler(None, aval)
    return lambda _, bufs: handler(*bufs)

def array_result_handler(aval):
  if not isinstance(aval._eltTy, BaseType): raise NotImplementedError
  padded_shape = []
  for d in aval.shape:
    if isinstance(d, int):
      padded_shape.append(d)
    elif isinstance(d, Var):
      padded_shape.append(d.aval._eltTy._bound)
    elif isinstance(d, DimIndexingExpr):
      padded_shape.append(d.name.aval._eltTy._bound)
    else:
      raise NotImplementedError  # TODO
  padded_aval = core.ShapedArray(tuple(padded_shape), aval._eltTy._dtype)
  array_handler = dispatch.array_result_handler(None, padded_aval)
  def handler(dims_dict, bufs):
    shape = tuple(dims_dict[d] if isinstance(d, Var) else
                  DimIndexer(dims_dict[d.name], d.indices) if isinstance(d, DimIndexingExpr) else
                  d for d in aval.shape)
    return Array(shape, aval._eltTy, array_handler(*bufs))
  return handler

def aval_to_num_buffers(aval):
  if isinstance(aval, AbsArray):
    return 1
  else:
    return len(xla.aval_to_xla_shapes(aval))

translations: Dict[core.Primitive, Callable] = {}


dynamic_xla_call_p = core.Primitive('dxla_call')
dynamic_xla_call_p.multiple_results = True

@dynamic_xla_call_p.def_impl
def _dynamic_xla_call_impl(*args, jaxpr, num_consts):
  in_dim_vals, consts, args = split_list(args, [len(jaxpr.in_dim_binders), num_consts])
  dim_in_avals = [v.aval for v in jaxpr.in_dim_binders]
  c = xc.XlaBuilder("dxla_call")
  dim_params, params = _make_params(c, dim_in_avals, map(xla.abstractify, args))
  const_params = _xla_consts(c, consts)
  dim_outs, outs = djaxpr_subcomp(c, jaxpr, dim_params, const_params + params)
  out = xops.Tuple(c, [o for ops in dim_outs + outs for o in ops])
  compiled = xb.get_backend(None).compile(c.build(out))
  result_handlers = map(result_handler, [v.aval for v in jaxpr.outs])
  out_bufcounts = [aval_to_num_buffers(v.aval) for v in jaxpr.outs]
  partitioner = result_partitioner(jaxpr.in_dim_binders, in_dim_vals,
                                   jaxpr.out_dims, out_bufcounts)
  return execute_compiled(compiled, partitioner, result_handlers,
                          in_dim_vals, args)

def djit(fun):
  def f_jitted(*args, **kwargs):
    args, in_tree = tree_flatten((args, kwargs))
    f, out_tree = flatten_fun(lu.wrap_init(fun), in_tree)
    # TODO we shouldn't dedup avals one array at a time; need to do it for the
    # full argument list!
    # unique_avals: Dict[int, core.AbstractValue] = {}
    in_avals = [core.raise_to_shaped(core.get_aval(x)) for x in args]
    jaxpr, consts, unconverted_binders = trace_to_jaxpr_dynamic(f, in_avals)
    num_consts = len(consts)
    args = [*consts, *args]
    dim_vals, args = _extract_dim_vals(jaxpr.in_dim_binders, jaxpr.in_binders,
                                       unconverted_binders, args)
    out_flat = dynamic_xla_call_p.bind(*dim_vals, *args, jaxpr=jaxpr,
                                       num_consts=num_consts)
    return tree_unflatten(out_tree(), out_flat)
  return f_jitted

def _extract_dim_vals(in_dim_binders, in_binders, unconverted_binders, args):
  converted_in_dim_vals, args = partition_list(unconverted_binders, args)
  sizes = {var: size for binder, arg in zip(in_binders, args)
           for var, size in zip(binder.aval.shape, np.shape(arg))
           if isinstance(var, Var)}
  num_binders = len(in_dim_binders) - len(converted_in_dim_vals)
  in_dim_vals = [sizes[v] for v in in_dim_binders[:num_binders]] + converted_in_dim_vals
  return in_dim_vals, args


def traceable_to_padded_translation(traceable):
  def translation(c, dims, avals, operands, **params):
    dim_avals = [core.ShapedArray((), np.int32) for _ in dims]
    padded_avals = map(_replace_vars_with_bounds, avals)

    @lu.wrap_init
    def fun(*args):
      dim_sizes, args = split_list(args, [len(dims)])
      logical_sizes = dict(zip(dims, dim_sizes))
      logical_shapes = [tuple([logical_sizes.get(d, d) for d in aval.shape])
                        for aval in avals]  # TODO more cases
      return traceable(logical_shapes, *args, **params)

    in_avals = [*dim_avals, *padded_avals]
    jaxpr, out_avals, consts = pe.trace_to_jaxpr_dynamic(fun, in_avals)

    operands_ = it.chain.from_iterable([*dims.values(), *operands])
    platform = "cpu"  # TODO: don't hardwire in the CPU translation.
    ctx = xla.TranslationContext(c, platform, xla.AxisEnv(1, (), ()), '')
    outs = xla.jaxpr_subcomp(ctx, jaxpr, xla._xla_consts(c, consts), *operands_)
    return util.unflatten(outs,
                          [aval_to_num_buffers(aval) for aval in out_avals])
  return translation

def _replace_vars_with_bounds(aval):
  if not isinstance(aval, AbsArray):
    return aval
  else:
    new_shape = []
    for d in aval.shape:
      if isinstance(d, Var):
        assert d.aval.shape == () and isinstance(d.aval._eltTy, BoundedIntTy)
        new_shape.append(d.aval._eltTy._bound)
      elif isinstance(d, int):
        new_shape.append(d)
      elif isinstance(d, BoundedInt):
        new_shape.append(d._bound)
      else:
        raise NotImplementedError(d)
    return core.ShapedArray(tuple(new_shape), aval._eltTy._dtype)

# AD

from jax.interpreters import ad

def _dynamic_xla_call_jvp(primals, tangents, *, jaxpr, num_consts):
  del num_consts
  in_dim_vals, primals = split_list(primals, [len(jaxpr.in_dim_binders)])
  _, tangents = split_list(tangents, [len(jaxpr.in_dim_binders)])
  new_jaxpr, consts = jvp_jaxpr(jaxpr)
  outs = dynamic_xla_call_p.bind(*in_dim_vals, *consts, *primals, *tangents,
                                 jaxpr=new_jaxpr, num_consts=len(consts))
  primals_out, tangents_out = split_list(outs, [len(outs) // 2])
  return primals_out, tangents_out
ad.primitive_jvps[dynamic_xla_call_p] = _dynamic_xla_call_jvp

def _dynamic_xla_call_transpose(cts_in, *args, jaxpr, num_consts):
  # TODO make this a dynamic_xla_call_p bind
  del num_consts
  vars_to_vals = dict(
      (d, t) for v, x in zip(jaxpr.in_binders, args)
      if isinstance(v.aval, AbsArray) and not ad.is_undefined_primal(x)
      for d, t in zip(v.aval.shape, x.shape) if isinstance(d, Var))
  dim_args = [vars_to_vals[v] for v in jaxpr.in_dim_binders]
  consts_bar, args_bar = backward_pass(jaxpr, dim_args, args, cts_in)  # type: ignore
  return [*consts_bar, *args_bar]
ad.primitive_transposes[dynamic_xla_call_p] = _dynamic_xla_call_transpose

def backward_pass(jaxpr, dim_args, args, cts_in):
  primal_env = {}
  ct_env = {}

  def write_cotangent(v, ct):
    ct_env[v] = ad.add_tangents(ct_env[v], ct) if v in ct_env else ct

  def read_cotangent(v):
    return ct_env.get(v, ad.Zero(v.aval))

  def read_primal(v):
    if type(v) is core.Literal:
      raise NotImplementedError  # TODO
    else:
      return primal_env.get(v, ad.UndefinedPrimal(v.aval))

  def write_primal(v, val):
    if not ad.is_undefined_primal(val):
      primal_env[v] = val

  write_primal(core.unitvar, core.unit)
  map(write_primal, jaxpr.in_dim_binders, dim_args)
  map(write_primal, jaxpr.in_binders, args)

  map(write_cotangent, jaxpr.outs, cts_in)
  raise NotImplementedError  # TODO finish this

def jvp_jaxpr(jaxpr):
  f = lu.wrap_init(jaxpr_as_fun(jaxpr))
  dimvars = dict((v, v.aval) for v in jaxpr.in_dim_binders)
  in_avals = [_replace_vars_with_avals(dimvars, v.aval) for v in jaxpr.in_binders]
  jaxpr, consts, _ = trace_to_jaxpr_dynamic(jvp_traceable(ad.jvp(f)), in_avals * 2)
  return jaxpr, consts

def _replace_vars_with_avals(dimvars, aval):
  if isinstance(aval, AbsArray):
    shape = [dimvars.get(d, d) for d in aval.shape]
    return AbsArray(tuple(shape), aval._eltTy)
  else:
    return aval

@lu.transformation
def jvp_traceable(*primals_and_tangents):
  n = len(primals_and_tangents)
  primals, tangents = split_list(primals_and_tangents, [n // 2])
  primals_out, tangents_out = yield (primals, tangents), {}
  yield (*primals_out, *tangents_out)

def _dynamic_xla_call_pe(trace, *tracers, jaxpr, num_consts):
  in_dim_tracers, tracers = split_list(tracers, [len(jaxpr.in_dim_binders)])
  if any(not t.pval.is_known() for t in in_dim_tracers):
    raise NotImplementedError
  in_unknowns = [not t.pval.is_known() for t in tracers]
  jaxpr1, jaxpr2, out_unknowns, num_res = partial_eval_jaxpr(jaxpr, in_unknowns)

  known_tracers, unknown_tracers = partition_list(in_unknowns, tracers)
  known_vals = [t.pval.get_known() for t in known_tracers]
  in_dim_vals = [t.pval.get_known() for t in in_dim_tracers]
  outs1_res = dynamic_xla_call_p.bind(*in_dim_vals, *known_vals, jaxpr=jaxpr1,
                                      num_consts=num_consts)
  outs1, res = split_list(outs1_res, [len(jaxpr1.outs) - num_res])

  in_dim_tracers = map(trace.new_instantiated_const, in_dim_tracers)
  res_tracers = map(trace.new_instantiated_const, res)
  outs2 = [pe.JaxprTracer(trace, pe.PartialVal.unknown(v.aval), None)
           for v in jaxpr2.outs]
  eqn = pe.new_eqn_recipe(in_dim_tracers + res_tracers + unknown_tracers, outs2,
                          dynamic_xla_call_p, dict(jaxpr=jaxpr2, num_consts=0),
                          source_info_util.new_source_info())
  for t in outs2: t.recipe = eqn
  outs1, outs2 = iter(outs1), iter(outs2)
  return [next(outs2) if uk else next(outs1) for uk in out_unknowns]
pe.custom_partial_eval_rules[dynamic_xla_call_p] = _dynamic_xla_call_pe

def partial_eval_jaxpr(jaxpr, in_unknowns):
  env: Dict[Var, bool] = {}
  res = []

  def read(v):
    if type(v) is core.Literal:
      raise NotImplementedError  # TODO
    else:
      return env[v]

  def write(unk, v):
    env[v] = unk

  def new_res(v):
    res.append(v)
    return v

  eqns1, eqns2 = [], []
  map(write, in_unknowns, jaxpr.in_binders)
  for eqn in jaxpr.eqns:
    unks = map(read, eqn.invars)
    if any(unks):
      invars = [v if unk else new_res(v) for unk, v in zip(unks, eqn.invars)]
      eqns2.append(pe.new_jaxpr_eqn(invars, eqn.outvars, eqn.primitive,
                                    eqn.params,
                                    source_info_util.new_source_info()))
      map(partial(write, True), eqn.outvars)
    else:
      eqns1.append(eqn)
      map(partial(write, False), eqn.outvars)
  out_unknowns = map(read, jaxpr.outs)
  out_dim_unknowns = map(read, jaxpr.out_dims)  # when linearizing, all known

  invars1, invars2 = partition_list(in_unknowns, jaxpr.in_binders)
  outvars1, outvars2 = partition_list(out_unknowns, jaxpr.outs)
  out_dims1, out_dims2 = partition_list(out_dim_unknowns, jaxpr.out_dims)

  outvars1 = outvars1 + res
  invars2 = res + invars2

  # TODO forward the correct residuals here (all dimvars used in types)
  in_dimvars2 = out_dims1 + jaxpr.in_dim_binders

  jaxpr1 = DJaxpr(jaxpr.in_dim_binders, invars1, out_dims1, outvars1, eqns1)
  jaxpr2 = DJaxpr(in_dimvars2,          invars2, out_dims2, outvars2, eqns2)

  return jaxpr1, jaxpr2, out_unknowns, len(res)


# batching

from jax.interpreters import batching

def _dynamic_xla_call_vmap(args, in_dims, *, jaxpr, num_consts):
  del num_consts
  in_dim_vals, args = split_list(args, [len(jaxpr.in_dim_binders)])
  in_dim_bdims, arg_bdims = split_list(in_dims, [len(jaxpr.in_dim_binders)])
  assert all(d is batching.not_mapped for d in in_dim_bdims)
  axis_size, = {x.shape[d] for x, d in zip(args, arg_bdims)
                if d is not batching.not_mapped}
  new_jaxpr, consts, out_dims = batch_jaxpr(jaxpr, axis_size, arg_bdims)
  outs = dynamic_xla_call_p.bind(*in_dim_vals, *consts, *args,
                                 jaxpr=new_jaxpr, num_consts=len(consts))
  return outs, out_dims
batching.primitive_batchers[dynamic_xla_call_p] = _dynamic_xla_call_vmap

def batch_jaxpr(jaxpr, axis_size, in_dims):
  dimvars = dict((v, v.aval) for v in jaxpr.in_dim_binders)
  in_avals = [_replace_vars_with_avals(dimvars, v.aval) for v in jaxpr.in_binders]

  in_avals = [core.unmapped_aval(axis_size, core.no_axis_name, d, aval)
              if d is not batching.not_mapped else aval
              for d, aval in zip(in_dims, in_avals)]

  fun, out_dims = batching.batch_subtrace(lu.wrap_init(jaxpr_as_fun(jaxpr)))
  f = _batch_fun(fun, in_dims)
  jaxpr, consts, _ = trace_to_jaxpr_dynamic(f, in_avals)
  return jaxpr, consts, out_dims()

@lu.transformation
def _batch_fun(in_dims, *in_vals, **params):
  with core.new_main(batching.BatchTrace, axis_name=core.no_axis_name) as main:
    out_vals = yield (main, in_dims, *in_vals), params
    del main
  yield out_vals

def _map_array(size: int, axis: int, aval: AbsArray) -> AbsArray:
  return AbsArray(tuple_delete(aval.shape, axis), aval._eltTy)

def _unmap_array(size: int, axis: int, aval: AbsArray) -> AbsArray:
  raise NotImplementedError

core.aval_mapping_handlers[AbsArray] = _map_array, _unmap_array


# Primitives

import numpy as np
from jax._src.lax import lax


## sin

def sin(x: Any) -> Any:
  return sin_p.bind(x)
sin_p = core.Primitive('sin_p')

@sin_p.def_abstract_eval
def _sin_abstract_eval(x):
  if isinstance(x, AbsArray):
    return AbsArray(x.shape, x._eltTy)
  else:
    return lax.sin_p.abstract_eval(x)

def _sin_typecheck_rule(invar):
  return [invar.aval]
typecheck_rules[sin_p] = _sin_typecheck_rule

def _sin_translation_rule(c, dims, avals, operands):
  (x,), = operands
  return [[xops.Sin(x)]]
translations[sin_p] = _sin_translation_rule

ad.defjvp(sin_p, lambda g, x: mul(g, cos(x)))


## cos

def cos(x: Any) -> Any:
  return cos_p.bind(x)
cos_p = core.Primitive('cos_p')

@cos_p.def_abstract_eval
def _cos_abstract_eval(x):
  if isinstance(x, AbsArray):
    return AbsArray(x.shape, x._eltTy)
  else:
    return lax.cos_p.abstract_eval(x)

def _cos_typecheck_rule(invar):
  return [invar.aval]
typecheck_rules[cos_p] = _cos_typecheck_rule

def _cos_translation_rule(c, dims, avals, operands):
  (x,), = operands
  return [[xops.Cos(x)]]
translations[cos_p] = _cos_translation_rule


## reduce-sum

def reduce_sum(x: Any, axes: Optional[Sequence[int]] = None) -> Any:
  if axes is None:
    axes = tuple(range(len(x.shape)))
  return reduce_sum_p.bind(x, axes=axes)
reduce_sum_p = core.Primitive('reduce_sum')

@reduce_sum_p.def_abstract_eval
def _sum_abstract_eval(operand, *, axes):
  if isinstance(operand, AbsArray):
    axes = set(axes)
    new_shape = [d for i, d in enumerate(operand.shape) if i not in axes]
    if (all(isinstance(d, int) for d in new_shape) and
        isinstance(operand._eltTy, BaseType)):
      return core.ShapedArray(tuple(new_shape), operand._eltTy._dtype)
    else:
      return AbsArray(tuple(new_shape), operand._eltTy)
  else:
    return lax.reduce_sum_p.reduce_sum_abstract_eval(operand, axes=axes)

def _reduce_sum_typecheck_rule(x, *, axes):
  return [reduce_sum_p.abstract_eval(x.aval, axes=axes)]
typecheck_rules[reduce_sum_p] = _reduce_sum_typecheck_rule

def _reduce_sum_translation_traceable(logical_shapes, x, *, axes):
  shape, = logical_shapes
  x = _replace_masked_values(shape, x, 0, axes=axes)
  return [lax._reduce_sum(x, axes=axes)]
translations[reduce_sum_p] = traceable_to_padded_translation(
    _reduce_sum_translation_traceable)

def _replace_masked_values(logical_shape, x, val, axes=None):
  axes = axes or set(range(len(logical_shape)))
  masks = [lax.broadcasted_iota(np.int32, x.shape, i) < d
           for i, d in enumerate(logical_shape) if d is not None and i in axes]
  if masks:
    x = lax.select(reduce(op.and_, masks), x, lax.full_like(x, val))
  return x

def _reduce_sum_transpose_rule(cotangent, operand, *, axes):
  raise NotImplementedError  # TODO
ad.deflinear2(reduce_sum_p, _reduce_sum_transpose_rule)


### lt

def lt(x, y):
  return lt_p.bind(x, y)
lt_p = core.Primitive('lt')

@lt_p.def_abstract_eval
def _lt_abstract_eval(x, y):
  if isinstance(x, AbsArray) or isinstance(y, AbsArray):
    # TODO check dtypes match
    if not x.shape:
      return AbsArray(y.shape, BaseType(np.dtype('bool')))
    if not y.shape:
      return AbsArray(x.shape, BaseType(np.dtype('bool')))
    map(_dims_must_equal, x.shape, y.shape)
    return AbsArray(x.shape, BaseType(np.dtype('bool')))
  else:
    return lax.lt_p.abstract_eval(x, y)

def _lt_typecheck_rule(x, y):
  return [lt_p.abstract_eval(x.aval, y.aval)]

def _lt_translation_rule(c, dims, avals, operands):
  (x,), (y,) = operands
  return [[xops.Lt(x, y)]]


### dot

def dot(x, y):
  assert len(x.shape) == len(y.shape) == 2
  return dot_general(x, y, ([1], [0]), ([], []))

Dims = Tuple[Sequence[int], Sequence[int]]

def dot_general(x: Any, y: Any, contract: Dims, batch: Dims) -> Any:
  return dot_general_p.bind(x, y, contract=contract, batch=batch)
dot_general_p = core.Primitive('dot_general')

@dot_general_p.def_abstract_eval
def _dot_general_abstract_eval(x, y, *, contract, batch):
  for i, j in zip(*contract): _dims_must_equal(x.shape[i], y.shape[j])
  for i, j in zip(*batch): _dims_must_equal(x.shape[i], y.shape[j])
  shape = lax._dot_general_shape_computation(x.shape, y.shape, (contract, batch))
  return AbsArray(shape, x._eltTy)

def _dot_general_typecheck_rule(x, y, *, contract, batch):
  return [_dot_general_abstract_eval(x.aval, y.aval,
                                     contract=contract, batch=batch)]
typecheck_rules[dot_general_p] = _dot_general_typecheck_rule

def _dot_general_trans(logical_shapes, x, y, *, contract, batch):
  x_shape, _ = logical_shapes
  lhs_contract, _ = contract
  x = _replace_masked_values(x_shape, x, 0, axes=lhs_contract)
  return [lax.dot_general(x, y, dimension_numbers=(contract, batch))]
translations[dot_general_p] = traceable_to_padded_translation(_dot_general_trans)

def _dot_general_transpose_rule(cotangent, x, y, *, contract, batch):
  assert False  # TODO
ad.primitive_transposes[dot_general_p] = _dot_general_transpose_rule


## add

def add(x: Any, y: Any) -> Any:
  return add_p.bind(x, y)
add_p = core.Primitive('add')

@add_p.def_abstract_eval
def _add_abstract_eval(x, y):
  if isinstance(x, AbsArray) and isinstance(y, AbsArray):
    map(_dims_must_equal, x.shape, y.shape)  # TODO broadcasting?
    return AbsArray(x.shape, x._eltTy)
  else:
    return lax.add_p.abstract_eval(x, y)

def _dims_must_equal(d1, d2):
  if isinstance(d1, (Tracer, Var)) and isinstance(d2, (Tracer, Var)):
    if d1.aval is d2.aval: return True
  elif isinstance(d1, int) and isinstance(d2, int):
    return d1 == d2
  raise Exception("can't prove shapes equal (or unequal)!")

def _add_typecheck_rule(x, y):
  return [add_p.abstract_eval(x.aval, y.aval)]
typecheck_rules[add_p] = _add_typecheck_rule

def _add_translation_rule(c, dims, avals, operands):
  (x,), (y,) = operands
  return [[xops.Add(x, y)]]
translations[add_p] = _add_translation_rule


## mul

def mul(x: Any, y: Any) -> Any:
  return mul_p.bind(x, y)
mul_p = core.Primitive('mul')

@mul_p.def_abstract_eval
def _mul_abstract_eval(x, y):
  if isinstance(x, AbsArray) and isinstance(y, AbsArray):
    map(_dims_must_equal, x.shape, y.shape)  # TODO broadcasting?
    return AbsArray(x.shape, x._eltTy)
  else:
    return lax.mul_p.abstract_eval(x, y)

def _mul_typecheck_rule(x, y):
  return [mul_p.abstract_eval(x.aval, y.aval)]
typecheck_rules[mul_p] = _mul_typecheck_rule

def _mul_translation_rule(c, dims, avals, operands):
  (x,), (y,) = operands
  return [[xops.Mul(x, y)]]
translations[mul_p] = _mul_translation_rule


## nonzero

def nonzero(x):
  return nonzero_p.bind(x)
nonzero_p = core.Primitive('nonzero')

def _nonzero_unpack_result(x):
  return [x.shape[-1], x]
nonzero_p.unpack_result = _nonzero_unpack_result  # type: ignore

def _nonzero_staging_rule(trace, tracers, params):
  aval = tracers[0].aval
  if isinstance(aval, AbsArray) and not isinstance(aval._eltTy, BaseType):
    raise NotImplementedError
  bound = aval.shape[-1]
  bound = bound if isinstance(bound, int) else bound._bound
  out_dim_aval = AbsArray(aval.shape[:-1], BoundedIntTy(bound))
  out_dim_tracer = pe.DynamicJaxprTracer(trace, out_dim_aval, None)
  if len(aval.shape) == 1:
    out_val_aval = AbsArray((out_dim_tracer,), BaseType(np.dtype('int32')))
  else:
    indices = tuple(range(len(aval.shape[:-1])))
    expr = DimIndexingExpr(out_dim_tracer, indices)
    out_val_aval = AbsArray((*aval.shape[:-1], expr),
                              BaseType(np.dtype('int32')))
  out_val_tracer = pe.DynamicJaxprTracer(trace, out_val_aval, None)
  invars = map(trace.getvar, tracers)
  outvars = map(trace.makevar, [out_dim_tracer, out_val_tracer])
  eqn = pe.new_jaxpr_eqn(invars, outvars, nonzero_p, {},
      source_info_util.new_source_info())
  trace.frame.eqns.append(eqn)
  return out_val_tracer
custom_staging_rules[nonzero_p] = _nonzero_staging_rule

def _nonzero_typecheck_rule(invar):
  bound = invar.aval.shape[-1]
  bound = bound if isinstance(bound, int) else bound._bound
  newvar = core.gensym()
  out_dim_var = newvar(AbsArray(invar.aval.shape[:-1], BoundedIntTy(bound)))
  if len(invar.aval.shape) == 1:
    out_val_aval = AbsArray((out_dim_var,), BaseType(np.dtype('int32')))
  else:
    indices = tuple(range(len(out_dim_var.aval.shape)))  # pytype: disable=attribute-error
    expr = DimIndexingExpr(out_dim_var, indices)
    out_val_aval = AbsArray((*invar.aval.shape[:-1], expr),
                              BaseType(np.dtype('int32')))
  return out_dim_var, out_val_aval
typecheck_rules[nonzero_p] = _nonzero_typecheck_rule

def _nonzero_translation_traceable(logical_shapes, x):
  shape, = logical_shapes
  assert shape
  x = _replace_masked_values(shape, x, 0)
  nonzero_indicators = x != 0
  last_axis = len(shape) - 1
  out_sizes = lax._reduce_sum(nonzero_indicators.astype(np.int32), [last_axis])
  iota = lax.broadcasted_iota(np.int32, x.shape, dimension=last_axis)
  _, idx = lax.sort_key_val(~nonzero_indicators, iota, dimension=last_axis)
  return out_sizes, idx
translations[nonzero_p] = traceable_to_padded_translation(
    _nonzero_translation_traceable)

def _nonzero_vmap_rule(args, in_dims):
  (x,), (d,) = args, in_dims
  if d != 0: raise NotImplementedError
  return nonzero_p.bind(x), 0
batching.primitive_batchers[nonzero_p] = _nonzero_vmap_rule


## iota

def iota(n):
  return iota_p.bind(n)
iota_p = core.Primitive('iota')

def _iota_staging_rule(trace, tracers, params):
  tracer, = tracers
  n = trace.get_const(tracer)
  if n is not None:
    if type(n) is not int: raise NotImplementedError  # TODO batched version?
    out_aval = core.ShapedArray((n,), np.dtype('int32'))
    out_tracer = pe.DynamicJaxprTracer(trace, out_aval, None)
    outvar = trace.makevar(out_tracer)
    eqn = pe.new_jaxpr_eqn([], [outvar], iota_p, dict(size=n),
        source_info_util.new_source_info())
  else:
    aval = tracer.aval
    if not isinstance(aval, AbsArray): raise TypeError
    if aval.shape:
      indices = tuple(range(len(aval.shape)))
      out_aval = AbsArray((*aval.shape, DimIndexingExpr(tracer, indices)),
                             BaseType(np.dtype('int32')))
    else:
      out_aval = AbsArray((tracer,), BaseType(np.dtype('int32')))
    out_tracer = pe.DynamicJaxprTracer(trace, out_aval, None)
    outvar = trace.makevar(out_tracer)
    invar = trace.getvar(tracer)
    eqn = pe.new_jaxpr_eqn([invar], [outvar], iota_p, {},
        source_info_util.new_source_info())
  trace.frame.eqns.append(eqn)
  return out_tracer
custom_staging_rules[iota_p] = _iota_staging_rule

def _iota_typecheck_rule(*invars, size=None):
  if size is not None:
    if invars: raise TypeError
    return [core.ShapedArray((size,), np.dtype('int32'))]
  else:
    invar, = invars
    if not invar.aval.shape:
      return [AbsArray((invar,), BaseType(np.dtype('int32')))]
    else:
      indices = tuple(range(len(invar.aval.shape)))
      return [AbsArray((*invar.aval.shape, DimIndexingExpr(invar, indices)),
                         BaseType(np.dtype('int32')))]
typecheck_rules[iota_p] = _iota_typecheck_rule

def _iota_translation_rule(c, dims, avals, operands, *, size=None):
  if size is None:
    aval, = avals
    size = aval._eltTy._bound
    shape = aval.shape
  else:
    shape = ()
  etype = xla.dtype_to_primitive_type(np.dtype('int32'))
  xla_shape = xc.Shape.array_shape(etype, (*shape, size))
  return [[xops.Iota(c, xla_shape, len(shape))]]
translations[iota_p] = _iota_translation_rule


## broadcast

def broadcast(x, d):
  return broadcast_p.bind(x, d)
broadcast_p = core.Primitive('broadcast')

def _broadcast_staging_rule(trace, tracers, params):
  x, d = tracers
  d_const = trace.get_const(d)
  if d_const is not None:
    raise NotImplementedError  # TODO
  else:
    aval = x.aval
    dtype = aval._eltTy._dtype if isinstance(aval, AbsArray) else aval.dtype
    out_aval = AbsArray((d, *x.shape), BaseType(dtype))
    out_tracer = pe.DynamicJaxprTracer(trace, out_aval, None)
    eqn = pe.new_jaxpr_eqn([trace.getvar(x), trace.getvar(d)],
                           [trace.makevar(out_tracer)], broadcast_p, {},
                           source_info_util.new_source_info())
    trace.frame.eqns.append(eqn)
    return out_tracer
custom_staging_rules[broadcast_p] = _broadcast_staging_rule

def _broadcast_typecheck_rule(x, d):
  aval = x.aval
  dtype = aval._eltTy._dtype if isinstance(aval, AbsArray) else aval.dtype
  return [AbsArray((d, *x.aval.shape), BaseType(dtype))]
typecheck_rules[broadcast_p] = _broadcast_typecheck_rule

def _broadcast_translation_rule(c, dims, avals, operands, *, size=None):
  (x,), (_,) = operands
  if size is None:
    _, aval = avals
    assert not aval.shape
    size = aval._eltTy._bound
  return [[xops.Broadcast(x, (size,))]]
translations[broadcast_p] = _broadcast_translation_rule


# Examples

import jax.numpy as jnp

def bbarray(bound_shape: Tuple[int, ...], x: NDArray):
  sizes: Dict[int, BoundedInt] = {}
  shape = tuple(sizes.setdefault(d, BoundedInt(d, bound))
                for d, bound in zip(x.shape, bound_shape))
  slices = tuple(slice(d) for d in x.shape)
  padded_x = jnp.ones(bound_shape, x.dtype).at[slices].set(x)
  return Array(shape, BaseType(x.dtype), padded_x)

def ones_like(x):
  if isinstance(x, Array):  # doesn't work with tracers
    return Array(x.shape, x._eltTy, jnp.ones_like(x._data))
  else:
    return jnp.ones_like(x)


if __name__ == '__main__':
  import jax
  jax.config.update('jax_platform_name', 'cpu')
  def p(s): print('\n--- ' + str(s))

  ## Staging and typechecking

  p('typecheck identity')
  def f(x):
    return x
  x = jnp.array([0, 1])
  jaxpr, _, _ = make_djaxpr(f, x)
  print(jaxpr)
  print(typecheck_jaxpr(jaxpr))

  p('typecheck sin')
  def f(x):
    return sin(x)
  x = bbarray((5,), jnp.arange(3.))
  jaxpr, _, _ = make_djaxpr(f, x)
  print(jaxpr)
  print(typecheck_jaxpr(jaxpr))

  p('typecheck sin-and-add')
  def f(x):
    y = sin(x)
    z = sin(y)
    return add(y, z)
  x = bbarray((5,), jnp.arange(3.))
  jaxpr, _, _ = make_djaxpr(f, x)
  print(jaxpr)
  print(typecheck_jaxpr(jaxpr))

  p('typecheck iota(3)')
  def f():  # type: ignore
    return iota(3)
  jaxpr, _, _ = make_djaxpr(f)
  print(jaxpr)
  print(typecheck_jaxpr(jaxpr))

  p('typecheck nonzero')
  def f(x):
    return nonzero(x)
  x = jnp.array([1, 0, -2, 0, 3, 0])
  jaxpr, _, _ = make_djaxpr(f, x)
  print(jaxpr)
  print(typecheck_jaxpr(jaxpr))

  p('typecheck sum-of-nonzero')
  def f(x):
    return reduce_sum(nonzero(x), tuple(range(len(x.shape))))
  x = jnp.array([1, 0, -2, 0, 3, 0])
  jaxpr, _, _ = make_djaxpr(f, x)
  print(jaxpr)
  print(typecheck_jaxpr(jaxpr))


  ## XLA lowering and execution

  @djit
  def f(x):
    nonzero_idx = nonzero(x)
    return reduce_sum(nonzero_idx)
  p('execute sum of nonzero indices')
  x = jnp.array([0, 1, 0, 1, 0, 1])
  print(f(x))
  print('should be', np.sum(np.nonzero(x)[0]))

  @djit
  def f(x):
    return nonzero(x)
  p('execute nonzero')
  x = jnp.array([0, 1, 0, 1, 0, 1])
  print(f(x))
  print('should be', np.nonzero(x)[0])

  @djit
  def f(i):
    return iota(i)
  p('execute iota')
  print(f(BoundedInt(3, 5)))
  print('should be', np.arange(3))

  @djit
  def f(x, n):
    y = nonzero(x)
    return broadcast(y, n)
  p('execute broadcast')
  x = np.arange(3)
  n = BoundedInt(4, 5)
  print(f(x, n))  # type: ignore
  print(f'should be\n{np.broadcast_to(np.nonzero(x)[0], (4, 2))}')


  ## ad

  @djit
  def f(x):
    y = sin(x)
    return reduce_sum(y, axes=(0,))
  x = bbarray((5,), jnp.arange(2.))
  p('basic jvp')
  z, z_dot = jax.jvp(f, (x,), (ones_like(x),))
  print(z, z_dot)


  p('basic linearize')
  _, f_lin = jax.linearize(f, x)
  print(f_lin(ones_like(x)))


  ## vmap

  @djit
  def f(x):
    return nonzero(x)
  p('vmap of nonzero')
  xs = jnp.array([[0, 1, 0, 1, 0, 1],
                  [1, 1, 1, 1, 0, 1]])
  print(jax.vmap(f)(xs))


  ## dot

  @djit
  def f(x):
    return dot(x, x)
  p('dot(x, x)')
  x = bbarray((4, 4), np.arange(9., dtype=np.float32).reshape(3, 3))
  print(f(x))
  y = np.arange(9.).reshape(3, 3)
  print(f'should be\n{np.dot(y, y)}')



"""Utility functions adopted from scipy.signal."""

import scipy.signal as osp_signal
import warnings

from jax._src.numpy import lax_numpy as jnp


def _triage_segments(window, nperseg, input_length):
  """
  Parses window and nperseg arguments for spectrogram and _spectral_helper.
  This is a helper function, not meant to be called externally.
  Parameters
  ----------
  window : string, tuple, or ndarray
      If window is specified by a string or tuple and nperseg is not
      specified, nperseg is set to the default of 256 and returns a window of
      that length.
      If instead the window is array_like and nperseg is not specified, then
      nperseg is set to the length of the window. A ValueError is raised if
      the user supplies both an array_like window and a value for nperseg but
      nperseg does not equal the length of the window.
  nperseg : int
      Length of each segment
  input_length: int
      Length of input signal, i.e. x.shape[-1]. Used to test for errors.
  Returns
  -------
  win : ndarray
      window. If function was called with string or tuple than this will hold
      the actual array used as a window.
  nperseg : int
      Length of each segment. If window is str or tuple, nperseg is set to
      256. If window is array_like, nperseg is set to the length of the
      6
      window.
  """
  # parse window; if array like, then set nperseg = win.shape
  if isinstance(window, (str, tuple)):
    # if nperseg not specified
    if nperseg is None:
      nperseg = 256  # then change to default
    if nperseg > input_length:
      warnings.warn(f'nperseg = {nperseg} is greater than input length '
                    f' = {input_length}, using nperseg = {nperseg}')
      nperseg = input_length
    win = jnp.array(osp_signal.get_window(window, nperseg))
  else:
    win = jnp.asarray(window)
    if len(win.shape) != 1:
      raise ValueError('window must be 1-D')
    if input_length < win.shape[-1]:
      raise ValueError('window is longer than input signal')
    if nperseg is None:
      nperseg = win.shape[0]
    elif nperseg is not None:
      if nperseg != win.shape[0]:
        raise ValueError("value specified for nperseg is different"
                          " from length of window")
  return win, nperseg


def _median_bias(n):
  """
  Returns the bias of the median of a set of periodograms relative to
  the mean.
  See Appendix B from [1]_ for details.
  Parameters
  ----------
  n : int
      Numbers of periodograms being averaged.
  Returns
  -------
  bias : float
      Calculated bias.
  References
  ----------
  .. [1] B. Allen, W.G. Anderson, P.R. Brady, D.A. Brown, J.D.E. Creighton.
          "FINDCHIRP: an algorithm for detection of gravitational waves from
          inspiraling compact binaries", Physical Review D 85, 2012,
          :arxiv:`gr-qc/0509116`
  """
  ii_2 = jnp.arange(2., n, 2)
  return 1 + jnp.sum(1. / (ii_2 + 1) - 1. / ii_2)
